# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2007       Douglas S. Blank
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2008       Raphael Ackerman
# Copyright (C) 2008       Brian G. Matherly
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"Import from CSV Spreadsheet"

import codecs
import csv

# ------------------------------------------------------------------------
#
# Set up logging
#
# ------------------------------------------------------------------------
import logging

# -------------------------------------------------------------------------
#
# Standard Python Modules
#
# -------------------------------------------------------------------------
import time
from io import TextIOWrapper

from gramps.gen.config import config

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale
from gramps.gen.datehandler import parser as _dp
from gramps.gen.db import DbTxn
from gramps.gen.display.place import displayer as place_displayer
from gramps.gen.errors import GrampsImportError as Error
from gramps.gen.lib import (
    Attribute,
    ChildRef,
    Citation,
    Event,
    EventRef,
    EventType,
    Family,
    FamilyRelType,
    Name,
    NameType,
    Note,
    NoteType,
    Person,
    Place,
    PlaceName,
    PlaceRef,
    PlaceType,
    Source,
    Surname,
    Tag,
)
from gramps.gen.lib.eventroletype import EventRoleType
from gramps.gen.utils.id import create_id
from gramps.gen.utils.libformatting import ImportInfo
from gramps.gen.utils.string import gender as gender_map
from gramps.gen.utils.unknown import create_explanation_note

LOG = logging.getLogger(".ImportCSV")

_ = glocale.translation.sgettext
ngettext = glocale.translation.ngettext  # else "nearby" comments are ignored

# -------------------------------------------------------------------------
#
# Support Functions
#
# -------------------------------------------------------------------------


def get_primary_event_ref_from_type(dbase, person, event_name):
    """
    >>> get_primary_event_ref_from_type(dbase, Person(), "Baptism"):
    """
    for ref in person.event_ref_list:
        if ref.get_role() == EventRoleType.PRIMARY:
            event = dbase.get_event_from_handle(ref.ref)
            if event and event.type.is_type(event_name):
                return ref
    return None


# -------------------------------------------------------------------------
#
# Support and main functions
#
# -------------------------------------------------------------------------
def rd(line_number, row, col, key, default=None):
    """Return Row data by column name"""
    if key in col:
        if col[key] >= len(row):
            LOG.warning("missing '%s, on line %d" % (key, line_number))
            return default
        retval = row[col[key]].strip()
        if retval == "":
            return default
        else:
            return retval
    else:
        return default


def importData(dbase, filename, user):
    """Function called by Gramps to import data on persons in CSV format."""
    if dbase.get_feature("skip-import-additions"):  # don't add source or tags
        parser = CSVParser(dbase, user, None)
    else:
        parser = CSVParser(
            dbase,
            user,
            (
                config.get("preferences.tag-on-import-format")
                if config.get("preferences.tag-on-import")
                else None
            ),
        )
    try:
        with open(filename, "rb") as filehandle:
            line = filehandle.read(3)
            if line == codecs.BOM_UTF8:
                filehandle.seek(0)
                filehandle = TextIOWrapper(
                    filehandle, encoding="utf_8_sig", errors="replace", newline=""
                )
            else:  # just open with OS encoding
                filehandle.seek(0)
                filehandle = TextIOWrapper(filehandle, errors="replace", newline="")
            msg = parser.parse(filehandle)
            if msg:
                user.notify_error(_("Bad references"), msg)
    except EnvironmentError as err:
        user.notify_error(_("%s could not be opened\n") % filename, str(err))
        return
    return ImportInfo({_("Results"): _("done")})


# -------------------------------------------------------------------------
#
# CSV Parser
#
# -------------------------------------------------------------------------
class CSVParser:
    """Class to read data in CSV format from a file object."""

    def __init__(self, dbase, user, default_tag_format=None):
        self.db = dbase
        self.user = user
        self.trans = None
        self.lineno = 0
        self.index = 0
        self.fam_count = 0
        self.indi_count = 0
        self.place_count = 0
        self.pref = {}  # person ref, internal to this sheet
        self.fref = {}  # family ref, internal to this sheet
        self.placeref = {}
        self.eventref = {}
        self.place_types = {}
        # Build reverse dictionary, name to type number
        for items in PlaceType().get_map().items():  # (0, 'Custom')
            self.place_types[items[1]] = items[0]
            self.place_types[items[1].lower()] = items[0]
            if _(items[1]) != items[1]:
                self.place_types[_(items[1])] = items[0]
        # Add custom types:
        for custom_type in self.db.get_place_types():
            self.place_types[custom_type] = 0
            self.place_types[custom_type.lower()] = 0
        column2label = {
            "surname": ("lastname", "last_name", "surname", _("surname"), _("Surname")),
            "firstname": (
                "firstname",
                "first_name",
                "given_name",
                "given",
                "given name",
                _("given name"),
                _("given"),
                _("Given"),
                _("Given name"),
            ),
            "callname": (
                "call name",
                _("Call name"),
                "callname",
                "call_name",
                "call",
                _("Call"),
                _("call"),
            ),
            "tag": ("tag", _("tag"), _("Tag")),
            "title": ("title", _("title"), _("title", "Person or Place")),
            "prefix": ("prefix", _("prefix"), _("Prefix")),
            "suffix": ("suffix", _("suffix"), _("Suffix")),
            "gender": ("gender", _("gender"), _("Gender")),
            "source": ("source", _("source"), _("Source")),
            "note": ("note", _("note"), _("Note")),
            "birthplace": (
                "birthplace",
                "birth_place",
                "birth place",
                _("birth place"),
                _("Birth place"),
            ),
            "birthplace_id": (
                "birthplaceid",
                "birth_place_id",
                "birth place id",
                _("birth place id"),
                "birthplace_id",
            ),
            "birthdate": ("birthdate", "birth_date", "birth date", _("birth date")),
            "birthsource": (
                "birthsource",
                "birth_source",
                "birth source",
                _("birth source"),
            ),
            "baptismplace": ("baptismplace", "baptism place", _("baptism place")),
            "baptismplace_id": (
                "baptismplaceid",
                "baptism place id",
                _("baptism place id"),
                "baptism_place_id",
                "baptismplace_id",
            ),
            "baptismdate": ("baptismdate", "baptism date", _("baptism date")),
            "baptismsource": ("baptismsource", "baptism source", _("baptism source")),
            "burialplace": ("burialplace", "burial place", _("burial place")),
            "burialplace_id": (
                "burialplaceid",
                "burial place id",
                _("burial place id"),
                "burial_place_id",
                "burialplace_id",
            ),
            "burialdate": ("burialdate", "burial date", _("burial date")),
            "burialsource": ("burialsource", "burial source", _("burial source")),
            "deathplace": (
                "deathplace",
                "death_place",
                "death place",
                _("death place"),
            ),
            "deathplace_id": (
                "deathplaceid",
                "death place id",
                _("death place id"),
                "death_place_id",
                "deathplace_id",
            ),
            "deathdate": ("deathdate", "death_date", "death date", _("death date")),
            "deathsource": (
                "deathsource",
                "death_source",
                "death source",
                _("death source"),
            ),
            "deathcause": (
                "deathcause",
                "death_cause",
                "death cause",
                _("death cause"),
            ),
            "grampsid": (_("Gramps ID"), "grampsid", "id", "gramps_id", "gramps id"),
            "person": ("person", _("person"), _("Person")),
            "occupationdescr": (
                "occupationdescr",
                _("occupationdescr"),
                _("Occupation description"),
            ),
            "occupationdate": (
                "occupationdate",
                _("occupationdate"),
                _("Occupation date"),
            ),
            "occupationplace": (
                "occupationplace",
                _("occupationplace"),
                _("Occupation place"),
            ),
            "occupationplace_id": (
                "occupationplace_id",
                _("occupationplace_id"),
                _("Occupation place id"),
            ),
            "occupationsource": (
                "occupationsource",
                _("occupationsource"),
                _("Occupation source"),
            ),
            "residencedate": ("residencedate", _("residencedate"), _("residence date")),
            "residenceplace": (
                "residenceplace",
                _("residenceplace"),
                _("residence place"),
            ),
            "residenceplace_id": (
                "residenceplace_id",
                _("residenceplace_id"),
                _("residence place id"),
            ),
            "residencesource": (
                "residencesource",
                _("residencesource"),
                _("residence source"),
            ),
            "attributetype": ("attributetype", _("attributetype"), _("attribute type")),
            "attributevalue": (
                "attributevalue",
                _("attributevalue"),
                _("attribute value"),
            ),
            "attributesource": (
                "attributesource",
                _("attributesource"),
                _("attribute source"),
            ),
            # ----------------------------------
            "child": ("child", _("child"), _("Child")),
            "family": ("family", _("family"), _("Family")),
            # ----------------------------------
            "wife": (
                "mother",
                _("mother"),
                _("Mother"),
                "wife",
                _("wife"),
                _("Wife"),
                "parent2",
                _("parent2"),
            ),
            "husband": (
                "father",
                _("father"),
                _("Father"),
                "husband",
                _("husband"),
                _("Husband"),
                "parent1",
                _("parent1"),
            ),
            "marriage": ("marriage", _("marriage"), _("Marriage")),
            "date": ("date", _("date"), _("Date")),
            "place": ("place", _("place"), _("Place")),
            "place_id": ("place id", "place_id", "placeid", _("place id")),
            "name": ("name", _("name"), _("Name")),
            "type": ("type", _("type"), _("Type")),
            "latitude": ("latitude", _("latitude")),
            "longitude": ("longitude", _("longitude")),
            "code": ("code", _("code"), _("Code")),
            "enclosed_by": (
                "enclosed by",
                _("enclosed by"),
                "enclosed_by",
                _("enclosed_by"),
                "enclosedby",
            ),
        }
        lab2col_dict = []
        for key in list(column2label.keys()):
            for val in column2label[key]:
                lab2col_dict.append((val.lower(), key))
        self.label2column = dict(lab2col_dict)
        if default_tag_format:
            name = time.strftime(default_tag_format)
            tag = self.db.get_tag_from_name(name)
            if tag:
                self.default_tag = tag
            else:
                self.default_tag = Tag()
                self.default_tag.set_name(name)
        else:
            self.default_tag = None

    def cleanup_column_name(self, column):
        """Handle column aliases for CSV spreadsheet import and SQL."""
        return self.label2column.get(column, column)

    def read_csv(self, filehandle):
        "Read the data from the file and return it as a list."
        my_dialect = config.get("csv.dialect")
        my_delimiter = config.get("csv.delimiter")
        try:
            if my_dialect == _("Custom"):
                data = [
                    [r.strip() for r in row]
                    for row in csv.reader(filehandle, delimiter=my_delimiter)
                ]
            else:
                data = [
                    [r.strip() for r in row]
                    for row in csv.reader(filehandle, dialect=my_dialect)
                ]
        except csv.Error as err:
            self.user.notify_error(
                _("format error: line %(line)d: %(zero)s")
                % {"line": csv.reader.line_num, "zero": err}
            )
            return None
        return data

    def lookup(self, type_, id_):
        """
        Return the object of type type_ with id id_ from db or previously
        stored value.
        """
        if id_ is None:
            return None
        if type_ == "family":
            if id_.startswith("[") and id_.endswith("]"):
                id_ = self.db.fid2user_format(id_[1:-1])
                db_lookup = self.db.get_family_from_gramps_id(id_)
                if db_lookup is None:
                    return self.lookup(type_, id_)
                else:
                    return db_lookup
            else:
                id_ = self.db.fid2user_format(id_)
                if id_.lower() in self.fref:
                    return self.fref[id_.lower()]
                else:
                    return None
        elif type_ == "person":
            if id_.startswith("[") and id_.endswith("]"):
                id_ = self.db.id2user_format(id_[1:-1])
                db_lookup = self.db.get_person_from_gramps_id(id_)
                if db_lookup is None:
                    return self.lookup(type_, id_)
                else:
                    return db_lookup
            else:
                id_ = self.db.id2user_format(id_)
                if id_.lower() in self.pref:
                    return self.pref[id_.lower()]
                else:
                    return None
        elif type_ == "place":
            if id_.startswith("[") and id_.endswith("]"):
                id_ = self.db.pid2user_format(id_[1:-1])
                db_lookup = self.db.get_place_from_gramps_id(id_)
                if db_lookup is None:
                    return self.lookup(type_, id_)
                else:
                    return db_lookup
            else:
                id_ = self.db.pid2user_format(id_)
                if id_.lower() in self.placeref:
                    return self.placeref[id_.lower()]
                else:
                    return None
        elif type_ == "event":
            if id_.startswith("[") and id_.endswith("]"):
                id_ = self.db.eid2user_format(id_[1:-1])
                db_lookup = self.db.get_event_from_gramps_id(id_)
                if db_lookup is None:
                    return self.lookup(type_, id_)
                else:
                    return db_lookup
            else:
                id_ = self.db.eid2user_format(id_)
                if id_.lower() in self.eventref:
                    return self.eventref[id_.lower()]
                else:
                    return None
        else:
            LOG.warning("invalid lookup type in CSV import: '%s'" % type_)
            return None

    def storeup(self, type_, id_, object_):
        "Store object object_ of type type_ in a dictionary under key id_."
        if id_.startswith("[") and id_.endswith("]"):
            id_ = id_[1:-1]
            # return # do not store gramps people; go look them up
        if type_ == "person":
            id_ = self.db.id2user_format(id_)
            self.pref[id_.lower()] = object_
        elif type_ == "family":
            id_ = self.db.fid2user_format(id_)
            self.fref[id_.lower()] = object_
        elif type_ == "place":
            id_ = self.db.pid2user_format(id_)
            self.placeref[id_.lower()] = object_
        else:
            LOG.warning("invalid storeup type in CSV import: '%s'" % type_)

    def parse(self, filehandle):
        """
        Prepare the database and parse the input file.

        :param filehandle: open file handle positioned at start of the file
        """
        progress_title = _("CSV Import")
        with self.user.progress(progress_title, _("Reading data..."), 1) as step:
            data = self.read_csv(filehandle)

        with self.user.progress(
            progress_title, _("Importing data..."), len(data)
        ) as step:
            tym = time.time()
            self.db.disable_signals()
            with DbTxn(_("CSV import"), self.db, batch=True) as self.trans:
                if self.default_tag and self.default_tag.handle is None:
                    self.db.add_tag(self.default_tag, self.trans)
                self._parse_csv_data(data, step)
                err_msg = self._check_refs()
            self.db.enable_signals()
            self.db.request_rebuild()
            tym = time.time() - tym
            # Translators: leave all/any {...} untranslated
            msg = ngettext(
                "Import Complete: {number_of} second",
                "Import Complete: {number_of} seconds",
                tym,
            ).format(number_of=tym)
            LOG.debug(msg)
            LOG.debug("New Families: %d" % self.fam_count)
            LOG.debug("New Individuals: %d" % self.indi_count)
            LOG.debug("New Places: %d" % self.place_count)
        return err_msg

    def _check_refs(self):
        """Check that forward cross references were satisfied"""
        txt = ""
        expl_note = create_explanation_note(self.db)
        for key in self.placeref:
            place = self.placeref[key]
            if place.name.value == _("Unknown"):
                txt = (", " + key) if txt else key
                place.add_note(expl_note.handle)
                self.db.commit_place(place, self.trans)
        if txt:
            self.db.commit_note(expl_note, self.trans, time.time())
            return _("The following IDs were referenced but not found:\n" + txt)
        else:
            return None

    def _parse_csv_data(self, data, step):
        """Parse each line of the input data and act accordingly."""
        self.lineno = 0
        self.index = 0
        self.fam_count = 0
        self.indi_count = 0
        self.place_count = 0
        self.pref = {}  # person ref, internal to this sheet
        self.fref = {}  # family ref, internal to this sheet
        self.placeref = {}
        self.eventref = {}
        header = None
        line_number = 0
        for row in data:
            step()
            line_number += 1
            if "".join(row) == "":  # no blanks are allowed inside a table
                header = None  # clear headers, ready for next "table"
                continue
            ######################################
            if header is None:
                header = [self.cleanup_column_name(r.lower()) for r in row]
                col = {}
                count = 0
                for key in header:
                    col[key] = count
                    count += 1
                continue
            # four different kinds of data: person, family, and marriage
            if ("marriage" in header) or ("husband" in header) or ("wife" in header):
                self._parse_marriage(line_number, row, col)
            elif "family" in header:
                self._parse_family(line_number, row, col)
            elif any(("surname" in header, "person" in header)):
                self._parse_person(line_number, row, col)
            elif "place" in header:
                self._parse_place(line_number, row, col)
            else:
                LOG.warning("ignoring line %d" % line_number)
        return None

    def _parse_marriage(self, line_number, row, col):
        "Parse the content of a Marriage,Husband,Wife line."
        marriage_ref = rd(line_number, row, col, "marriage")
        husband = rd(line_number, row, col, "husband")
        wife = rd(line_number, row, col, "wife")
        marriagedate = rd(line_number, row, col, "date")
        marriageplace = rd(line_number, row, col, "place")
        marriageplace_id = rd(line_number, row, col, "place_id")
        marriagesource = rd(line_number, row, col, "source")
        note = rd(line_number, row, col, "note")
        tag = rd(line_number, row, col, "tag")
        wife = self.lookup("person", wife)
        husband = self.lookup("person", husband)

        if (
            husband is None
            and wife is None
            and (tag is not None or note is not None)
            and marriage_ref is not None
        ):
            # Just adding a note or tag to a marriage event
            marriage = self.lookup("event", marriage_ref)
            if marriage is None:
                LOG.warning("no marriage found for ref %s" % marriage_ref)
                return
            if tag:
                self.add_tag(marriage, tag)
            if note:
                self.add_note(NoteType.EVENT, marriage, note)
            self.db.commit_event(marriage, self.trans)
            return

        # Below, we assume editing/creating a family

        if husband is None and wife is None and marriage_ref is None:
            # might have children, so go ahead and add
            LOG.warning("no parents on line %d; adding family anyway" % line_number)

        family = self.get_or_create_family(marriage_ref, husband, wife)
        # adjust gender, if not already provided
        if husband:
            # this is just a guess, if unknown
            if husband.get_gender() == Person.UNKNOWN:
                husband.set_gender(Person.MALE)
                self.db.commit_person(husband, self.trans)
        if wife:
            # this is just a guess, if unknown
            if wife.get_gender() == Person.UNKNOWN:
                wife.set_gender(Person.FEMALE)
                self.db.commit_person(wife, self.trans)
        if marriage_ref:
            self.storeup("family", marriage_ref, family)
        if marriagesource:
            # add, if new
            new, marriagesource = self.get_or_create_source(marriagesource)
        if marriageplace and marriageplace_id:
            raise Error("Error in marriage: can't have a place and place_id")
        if marriageplace:
            # add, if new
            new, marriageplace = self.get_or_create_place(marriageplace)
        elif marriageplace_id:
            # better exist already, locally or in database:
            marriageplace = self.lookup("place", marriageplace_id)
        if marriagedate:
            marriagedate = _dp.parse(marriagedate)
        if marriagedate or marriageplace or marriagesource:
            # add, if new; replace, if different
            new, marriage = self.get_or_create_event(
                family, EventType.MARRIAGE, marriagedate, marriageplace, marriagesource
            )
            if new:
                mar_ref = EventRef()
                mar_ref.set_reference_handle(marriage.get_handle())
                mar_ref.set_role(EventRoleType(EventRoleType.FAMILY))
                family.add_event_ref(mar_ref)
            if tag:
                self.add_tag(marriage, tag)
            if note:
                self.add_note(NoteType.EVENT, marriage, note)
            self.db.commit_event(marriage, self.trans)

        self.db.commit_family(family, self.trans)

    def _parse_family(self, line_number, row, col):
        "Parse the content of a family line"
        family_ref = rd(line_number, row, col, "family")
        if family_ref is None:
            LOG.warning("no family reference found for family on line %d" % line_number)
            return  # required
        source = rd(line_number, row, col, "source")
        note = rd(line_number, row, col, "note")
        gender = rd(line_number, row, col, "gender")
        tag = rd(line_number, row, col, "tag")
        child_ref = rd(line_number, row, col, "child")
        child = self.lookup("person", child_ref)
        family = self.lookup("family", family_ref)
        if family is None:
            LOG.warning(
                "no matching family reference found for family "
                "on line %d" % line_number
            )
            return
        if child:
            # is this child already in this family? If so, don't add
            LOG.debug("children: %s", [ref.ref for ref in family.get_child_ref_list()])
            LOG.debug("looking for: %s", child.get_handle())
            if child.get_handle() not in [
                ref.ref for ref in family.get_child_ref_list()
            ]:
                # add child to family
                LOG.debug(
                    "   adding child [%s] to family [%s]",
                    child.get_gramps_id(),
                    family.get_gramps_id(),
                )
                childref = ChildRef()
                childref.set_reference_handle(child.get_handle())
                family.add_child_ref(childref)
                self.db.commit_family(family, self.trans)
                child.add_parent_family_handle(family.get_handle())
            if gender:
                # replace
                gender = gender.lower()
                if gender == gender_map[Person.MALE].lower():
                    gender = Person.MALE
                elif gender == gender_map[Person.FEMALE].lower():
                    gender = Person.FEMALE
                elif gender == gender_map[Person.OTHER].lower():
                    gender = Person.OTHER
                else:
                    gender = Person.UNKNOWN
                child.set_gender(gender)
            if source:
                dummy_new, source = self.get_or_create_source(source)
                self.find_and_set_citation(child, source)
            if tag is not None:
                self.add_tag(child, tag)
            if note:
                self.add_note(NoteType.PERSON, child, note)

            self.db.commit_person(child, self.trans)

        else:  # not a child; note and tag refers to family
            if tag is not None:
                self.add_tag(family, tag)
            if note:
                self.add_note(NoteType.EVENT, family, note)

        self.db.commit_family(family, self.trans)

    def _parse_person(self, line_number, row, col):
        "Parse the content of a Person line."
        surname = rd(line_number, row, col, "surname")
        firstname = rd(line_number, row, col, "firstname")
        callname = rd(line_number, row, col, "callname")
        tag = rd(line_number, row, col, "tag")
        title = rd(line_number, row, col, "title")
        prefix = rd(line_number, row, col, "prefix")
        suffix = rd(line_number, row, col, "suffix")
        gender = rd(line_number, row, col, "gender")
        source = rd(line_number, row, col, "source")
        note = rd(line_number, row, col, "note")
        birthplace = rd(line_number, row, col, "birthplace")
        birthplace_id = rd(line_number, row, col, "birthplace_id")
        birthdate = rd(line_number, row, col, "birthdate")
        birthsource = rd(line_number, row, col, "birthsource")
        baptismplace = rd(line_number, row, col, "baptismplace")
        baptismplace_id = rd(line_number, row, col, "baptismplace_id")
        baptismdate = rd(line_number, row, col, "baptismdate")
        baptismsource = rd(line_number, row, col, "baptismsource")
        burialplace = rd(line_number, row, col, "burialplace")
        burialplace_id = rd(line_number, row, col, "burialplace_id")
        burialdate = rd(line_number, row, col, "burialdate")
        burialsource = rd(line_number, row, col, "burialsource")
        deathplace = rd(line_number, row, col, "deathplace")
        deathplace_id = rd(line_number, row, col, "deathplace_id")
        deathdate = rd(line_number, row, col, "deathdate")
        deathsource = rd(line_number, row, col, "deathsource")
        deathcause = rd(line_number, row, col, "deathcause")
        grampsid = rd(line_number, row, col, "grampsid")
        person_ref = rd(line_number, row, col, "person")
        occupationdescr = rd(line_number, row, col, "occupationdescr")
        occupationplace = rd(line_number, row, col, "occupationplace")
        occupationplace_id = rd(line_number, row, col, "occupationplace_id")
        occupationsource = rd(line_number, row, col, "occupationsource")
        occupationdate = rd(line_number, row, col, "occupationdate")
        residencedate = rd(line_number, row, col, "residencedate")
        residenceplace = rd(line_number, row, col, "residenceplace")
        residenceplace_id = rd(line_number, row, col, "residenceplace_id")
        residencesource = rd(line_number, row, col, "residencesource")
        attributetype = rd(line_number, row, col, "attributetype")
        attributevalue = rd(line_number, row, col, "attributevalue")
        attributesource = rd(line_number, row, col, "attributesource")

        #########################################################
        # if this person already exists, don't create them
        person = self.lookup("person", person_ref)
        if person is None:
            if surname is None:
                LOG.warning("empty surname for new person on line %d" % line_number)
                surname = ""
            # new person
            person = self.create_person()
            name = Name()
            name.set_type(NameType(NameType.BIRTH))
            if firstname is not None:
                name.set_first_name(firstname)
            surname_obj = Surname()
            if surname is not None:
                surname_obj.set_surname(surname)
            name.add_surname(surname_obj)
            person.set_primary_name(name)
        else:
            name = person.get_primary_name()
        #########################################################
        if person_ref is not None:
            self.storeup("person", person_ref, person)
        # replace
        if surname is not None:
            name.get_primary_surname().set_surname(surname)
        if firstname is not None:
            name.set_first_name(firstname)
        if callname is not None:
            name.set_call_name(callname)
        if title is not None:
            name.set_title(title)
        if prefix is not None:
            name.get_primary_surname().set_prefix(prefix)
            name.group_as = ""  # HELP? what should I do here?
        if suffix is not None:
            name.set_suffix(suffix)
        if note is not None:
            # append notes, if previous notes
            previous_notes_list = person.get_note_list()
            updated_note = False
            for note_handle in previous_notes_list:
                previous_note = self.db.get_note_from_handle(note_handle)
                if previous_note.type == NoteType.PERSON:
                    previous_text = previous_note.get()
                    if note not in previous_text:
                        note = previous_text + "\n" + note
                    previous_note.set(note)
                    self.db.commit_note(previous_note, self.trans)
                    updated_note = True
                    break
            if not updated_note:
                # add new note here
                new_note = Note()
                new_note.handle = create_id()
                new_note.type.set(NoteType.PERSON)
                new_note.set(note)
                if self.default_tag:
                    new_note.add_tag(self.default_tag.handle)
                self.db.add_note(new_note, self.trans)
                person.add_note(new_note.handle)
        if grampsid is not None:
            person.gramps_id = grampsid
        elif person_ref is not None:
            if person_ref.startswith("[") and person_ref.endswith("]"):
                person.gramps_id = self.db.id2user_format(person_ref[1:-1])
        if person.get_gender() == Person.UNKNOWN and gender is not None:
            gender = gender.lower()
            if gender == gender_map[Person.MALE].lower():
                gender = Person.MALE
            elif gender == gender_map[Person.FEMALE].lower():
                gender = Person.FEMALE
            elif gender == gender_map[Person.OTHER].lower():
                gender = Person.OTHER
            else:
                gender = Person.UNKNOWN
            person.set_gender(gender)
        if tag is not None:
            self.add_tag(person, tag)
        #########################################################
        # add if new, replace if different
        # Birth:
        if birthdate is not None:
            birthdate = _dp.parse(birthdate)
        if birthplace and birthplace_id:
            raise Error("Error in person: can't have a birthplace and birthplace_id")
        if birthplace is not None:
            new, birthplace = self.get_or_create_place(birthplace)
        elif birthplace_id:
            # better exist already, locally or in database:
            birthplace = self.lookup("place", birthplace_id)
        if birthsource is not None:
            new, birthsource = self.get_or_create_source(birthsource)
        if birthdate or birthplace or birthsource:
            new, birth = self.get_or_create_event(
                person, EventType.BIRTH, birthdate, birthplace, birthsource
            )
            birth_ref = person.get_birth_ref()
            if birth_ref is None:
                # new
                birth_ref = EventRef()
            birth_ref.set_reference_handle(birth.get_handle())
            person.set_birth_ref(birth_ref)
        # Baptism:
        if baptismdate is not None:
            baptismdate = _dp.parse(baptismdate)
        if baptismplace and baptismplace_id:
            raise Error(
                "Error in person: can't have a baptismplace and baptismplace_id"
            )
        if baptismplace is not None:
            new, baptismplace = self.get_or_create_place(baptismplace)
        elif baptismplace_id:
            # better exist already, locally or in database:
            baptismplace = self.lookup("place", baptismplace_id)
        if baptismsource is not None:
            new, baptismsource = self.get_or_create_source(baptismsource)
        if baptismdate or baptismplace or baptismsource:
            new, baptism = self.get_or_create_event(
                person, EventType.BAPTISM, baptismdate, baptismplace, baptismsource
            )
            baptism_ref = get_primary_event_ref_from_type(self.db, person, "Baptism")
            if baptism_ref is None:
                # new
                baptism_ref = EventRef()
            baptism_ref.set_reference_handle(baptism.get_handle())
            person.add_event_ref(baptism_ref)
        # Death:
        if deathdate is not None:
            deathdate = _dp.parse(deathdate)
        if deathplace and deathplace_id:
            raise Error("Error in person: can't have a deathplace and deathplace_id")
        if deathplace is not None:
            new, deathplace = self.get_or_create_place(deathplace)
        elif deathplace_id:
            # better exist already, locally or in database:
            deathplace = self.lookup("place", deathplace_id)
        if deathsource is not None:
            new, deathsource = self.get_or_create_source(deathsource)
        if deathdate or deathplace or deathsource or deathcause:
            new, death = self.get_or_create_event(
                person, EventType.DEATH, deathdate, deathplace, deathsource
            )
            if deathcause:
                death.set_description(deathcause)
                self.db.commit_event(death, self.trans)
            death_ref = person.get_death_ref()
            if death_ref is None:
                # new
                death_ref = EventRef()
            death_ref.set_reference_handle(death.get_handle())
            person.set_death_ref(death_ref)
        # Burial:
        if burialdate is not None:
            burialdate = _dp.parse(burialdate)
        if burialplace and burialplace_id:
            raise Error("Error in person: can't have a burialplace and burialplace_id")
        if burialplace is not None:
            new, burialplace = self.get_or_create_place(burialplace)
        elif burialplace_id:
            # better exist already, locally or in database:
            burialplace = self.lookup("place", burialplace_id)
        if burialsource is not None:
            new, burialsource = self.get_or_create_source(burialsource)
        if burialdate or burialplace or burialsource:
            new, burial = self.get_or_create_event(
                person, EventType.BURIAL, burialdate, burialplace, burialsource
            )
            burial_ref = get_primary_event_ref_from_type(self.db, person, "Burial")
            if burial_ref is None:
                # new
                burial_ref = EventRef()
            burial_ref.set_reference_handle(burial.get_handle())
            person.add_event_ref(burial_ref)
        if source:
            # add, if new
            new, source = self.get_or_create_source(source)
            self.find_and_set_citation(person, source)

        # Attribute
        # update existing custom attribute or create it
        if attributevalue is not None:
            new, attr = self.get_or_create_attribute(
                person, attributetype, attributevalue, attributesource
            )

        # Occupation:
        # Contrary to the fields above,
        # each line in the csv will add a new occupation event
        if occupationdescr is not None:  # if no description we have no info to add
            if occupationdate is not None:
                occupationdate = _dp.parse(occupationdate)
            # occupation place takes precedence over place id if both are set
            if occupationplace is not None:
                new, occupationplace = self.get_or_create_place(occupationplace)
            elif occupationplace_id:
                occupationplace = self.lookup("place", occupationplace_id)
            if occupationsource is not None:
                new, occupationsource = self.get_or_create_source(occupationsource)
            new, occupation = self.get_or_create_event(
                person,
                EventType.OCCUPATION,
                occupationdate,
                occupationplace,
                occupationsource,
                occupationdescr,
                True,
            )
            occupation_ref = EventRef()
            occupation_ref.set_reference_handle(occupation.get_handle())
            person.add_event_ref(occupation_ref)

        # Residence:
        # Contrary to the fields above occupation,
        # each line in the csv will add a new residence event
        if residencedate is not None:
            residencedate = _dp.parse(residencedate)
        # residence  place takes precedence over place id if both are set
        if residenceplace is not None:
            new, residenceplace = self.get_or_create_place(residenceplace)
        elif residenceplace_id:
            residenceplace = self.lookup("place", residenceplace_id)
        if residencesource is not None:
            new, residencesource = self.get_or_create_source(residencesource)
        if residencedate or residenceplace or residencesource:
            new, residence = self.get_or_create_event(
                person,
                EventType.RESIDENCE,
                residencedate,
                residenceplace,
                residencesource,
                None,
                True,
            )
            residence_ref = EventRef()
            residence_ref.set_reference_handle(residence.get_handle())
            person.add_event_ref(residence_ref)

        self.db.commit_person(person, self.trans)

    def _parse_place(self, line_number, row, col):
        "Parse the content of a Place line."
        place_id = rd(line_number, row, col, "place")
        place_title = rd(line_number, row, col, "title")
        place_name = rd(line_number, row, col, "name")
        place_type_str = rd(line_number, row, col, "type")
        place_latitude = rd(line_number, row, col, "latitude")
        place_longitude = rd(line_number, row, col, "longitude")
        place_code = rd(line_number, row, col, "code")
        place_enclosed_by_id = rd(line_number, row, col, "enclosed_by")
        place_date = rd(line_number, row, col, "date")
        tag = rd(line_number, row, col, "tag")
        #########################################################
        # if this place already exists, don't create it
        place = self.lookup("place", place_id)
        if place is None:
            # new place
            place = self.create_place()
            if place_id is not None:
                if place_id.startswith("[") and place_id.endswith("]"):
                    place.gramps_id = self.db.pid2user_format(place_id[1:-1])
                self.storeup("place", place_id, place)
        if place_title is not None:
            place.title = place_title
        if tag is not None:
            self.add_tag(place, tag)
        if place_name is not None:
            place.name = PlaceName(value=place_name)
        if place_type_str is not None:
            place.place_type = self.get_place_type(place_type_str)
        if place_latitude is not None:
            place.lat = place_latitude
        if place_longitude is not None:
            place.long = place_longitude
        if place_code is not None:
            place.code = place_code
        if place_enclosed_by_id is not None:
            place_enclosed_by = self.lookup("place", place_enclosed_by_id)
            if place_enclosed_by is None:
                # Not yet found in import, so store for later
                place_enclosed_by = self.create_place()
                place_enclosed_by.name.set_value(_("Unknown"))
                if place_enclosed_by_id.startswith(
                    "["
                ) and place_enclosed_by_id.endswith("]"):
                    place_enclosed_by.gramps_id = self.db.pid2user_format(
                        place_enclosed_by_id[1:-1]
                    )
                self.storeup("place", place_enclosed_by_id, place_enclosed_by)
            for placeref in place.placeref_list:
                if place_enclosed_by.handle == placeref.ref:
                    break
            else:
                placeref = PlaceRef()
                placeref.ref = place_enclosed_by.handle
                place.placeref_list.append(placeref)
            if place_date:
                placeref.date = _dp.parse(place_date)
        #########################################################
        self.db.commit_place(place, self.trans)

    def get_place_type(self, place_type_str):
        if place_type_str in self.place_types:
            return PlaceType((self.place_types[place_type_str], place_type_str))
        else:
            # New custom type:
            return PlaceType((0, place_type_str))

    def get_or_create_family(self, family_ref, husband, wife):
        "Return the family object for the give family ID."
        # if a gramps_id and exists:
        LOG.debug("get_or_create_family")
        if family_ref.startswith("[") and family_ref.endswith("]"):
            id_ = self.db.fid2user_format(family_ref[1:-1])
            family = self.db.get_family_from_gramps_id(id_)
            if family:
                # don't delete, only add
                fam_husband_handle = family.get_father_handle()
                fam_wife_handle = family.get_mother_handle()
                if husband:
                    if husband.get_handle() != fam_husband_handle:
                        # this husband is not the same old one! Add him!
                        family.set_father_handle(husband.get_handle())
                if wife:
                    if wife.get_handle() != fam_wife_handle:
                        # this wife is not the same old one! Add her!
                        family.set_mother_handle(wife.get_handle())
                LOG.debug("   returning existing family")
                return family
        # if not, create one:
        family = Family()
        # was marked with a gramps_id, but didn't exist, so we'll use it:
        if family_ref.startswith("[") and family_ref.endswith("]"):
            id_ = self.db.fid2user_format(family_ref[1:-1])
            family.set_gramps_id(id_)
        # add it:
        family.set_handle(create_id())
        if self.default_tag:
            family.add_tag(self.default_tag.handle)
        if husband:
            family.set_father_handle(husband.get_handle())
            husband.add_family_handle(family.get_handle())
        if wife:
            family.set_mother_handle(wife.get_handle())
            wife.add_family_handle(family.get_handle())
        if husband and wife:
            family.set_relationship(FamilyRelType.MARRIED)
        self.db.add_family(family, self.trans)
        if husband:
            self.db.commit_person(husband, self.trans)
        if wife:
            self.db.commit_person(wife, self.trans)
        self.fam_count += 1
        return family

    def get_or_create_event(
        self,
        object_,
        type_,
        date=None,
        place=None,
        source=None,
        descr=None,
        create_only=False,
    ):
        # first, see if it exists
        LOG.debug("get_or_create_event")
        ref_list = object_.get_event_ref_list()
        LOG.debug("refs: %s", ref_list)
        # look for a match, and possible correction
        # except if create_only is true (for events that
        # can have several occurrences like occupations, residences)
        if not create_only:
            for ref in ref_list:
                event = self.db.get_event_from_handle(ref.ref)
                LOG.debug(
                    "   compare event type %s == %s", int(event.get_type()), type_
                )
                if int(event.get_type()) == type_:
                    # Match! Let's update
                    if date:
                        event.set_date_object(date)
                    if place:
                        event.set_place_handle(place.get_handle())
                    if source:
                        self.find_and_set_citation(event, source)
                    if descr:
                        event.set_description(descr)
                    self.db.commit_event(event, self.trans)
                    LOG.debug("   returning existing event")
                    return (0, event)
        # else create it:
        LOG.debug("   creating event")
        event = Event()
        if type_:
            event.set_type(EventType(type_))
        if date:
            event.set_date_object(date)
        if place:
            event.set_place_handle(place.get_handle())
        if source:
            self.find_and_set_citation(event, source)
        if descr:
            event.set_description(descr)
        if self.default_tag:
            event.add_tag(self.default_tag.handle)
        self.db.add_event(event, self.trans)
        return (1, event)

    def get_or_create_attribute(self, object_, type_, value_, source=None):
        "Replaces existing attribute or create it"
        LOG.debug("get_or_create_attribute")
        attr_list = object_.get_attribute_list()
        LOG.debug("refs: %s", attr_list)
        # remove attributes if it already exists
        if type_ is None:
            type_ = "UNKNOWN"
        for attr in attr_list:
            if attr.get_type() == type_:
                object_.remove_attribute(attr)
        # then add it
        LOG.debug("adding attribute")
        attr = Attribute()
        attr.set_type(type_)
        attr.set_value(value_)
        if source is not None:
            new, source = self.get_or_create_source(source)
            self.find_and_set_citation(attr, source)
        object_.add_attribute(attr)
        return (1, attr)

    def create_person(self):
        """Used to create a new person we know doesn't exist"""
        person = Person()
        if self.default_tag:
            person.add_tag(self.default_tag.handle)
        self.db.add_person(person, self.trans)
        self.indi_count += 1
        return person

    def create_place(self):
        """Used to create a new person we know doesn't exist"""
        place = Place()
        if self.default_tag:
            place.add_tag(self.default_tag.handle)
        self.db.add_place(place, self.trans)
        self.place_count += 1
        return place

    def get_or_create_place(self, place_name):
        "Return the requested place object tuple-packed with a new indicator."
        if place_name.startswith("[") and place_name.endswith("]"):
            place = self.lookup("place", place_name)
            return (0, place)
        LOG.debug("get_or_create_place: looking for: %s", place_name)
        for place_handle in self.db.iter_place_handles():
            place = self.db.get_place_from_handle(place_handle)
            place_title = place_displayer.display(self.db, place)
            if place_title == place_name:
                return (0, place)
        place = Place()
        place.set_title(place_name)
        place.name = PlaceName(value=place_name)
        self.db.add_place(place, self.trans)
        self.place_count += 1
        return (1, place)

    def get_or_create_source(self, source_text):
        "Return the requested source object tuple-packed with a new indicator."
        source_list = self.db.get_source_handles(sort_handles=False)
        LOG.debug("get_or_create_source: list: %s", source_list)
        LOG.debug("get_or_create_source: looking for: %s", source_text)
        for source_handle in source_list:
            source = self.db.get_source_from_handle(source_handle)
            if source.get_title() == source_text:
                LOG.debug("   returning existing source")
                return (0, source)
        LOG.debug("   creating source")
        source = Source()
        source.set_title(source_text)
        self.db.add_source(source, self.trans)
        return (1, source)

    def find_and_set_citation(self, obj, source):
        # look for the source in the existing citations for the object
        LOG.debug(
            "find_and_set_citation: looking for source: %s", source.get_gramps_id()
        )
        for citation_handle in obj.get_citation_list():
            citation = self.db.get_citation_from_handle(citation_handle)
            LOG.debug(
                "find_and_set_citation: existing citation: %s", citation.get_gramps_id()
            )
            poss_source = self.db.get_source_from_handle(
                citation.get_reference_handle()
            )
            LOG.debug(
                "   compare source %s == %s",
                source.get_gramps_id(),
                poss_source.get_gramps_id(),
            )
            if poss_source.get_handle() == source.get_handle():
                # The source is already cited
                LOG.debug("   source already cited")
                return
        # we couldn't find an appropriate citation, so we have to create one.
        citation = Citation()
        LOG.debug("   creating citation")
        citation.set_reference_handle(source.get_handle())
        self.db.add_citation(citation, self.trans)
        LOG.debug(
            "   created citation, citation %s %s" % (citation, citation.get_gramps_id())
        )
        obj.add_citation(citation.get_handle())

    def add_tag(self, obj, name):
        """
        Add a tag to an obj and database if it doesn't have it already
        """
        tag = self.db.get_tag_from_name(name)
        if tag is None:
            tag = Tag()
            tag.set_name(name)
            self.db.add_tag(tag, self.trans)
        obj.add_tag(tag.handle)

    def add_note(self, note_type, obj, note):
        """
        Add a note to an object
        """
        previous_notes_list = obj.get_note_list()
        updated_note = False
        for note_handle in previous_notes_list:
            previous_note = self.db.get_note_from_handle(note_handle)
            if previous_note.type == note_type:
                previous_text = previous_note.get()
                if note not in previous_text:
                    note = previous_text + "\n" + note
                previous_note.set(note)
                self.db.commit_note(previous_note, self.trans)
                updated_note = True
                break
        if not updated_note:
            new_note = Note()
            new_note.handle = create_id()
            new_note.type.set(note_type)
            new_note.set(note)
            if self.default_tag:
                new_note.add_tag(self.default_tag.handle)
            self.db.add_note(new_note, self.trans)
            obj.add_note(new_note.handle)

        if note_type == NoteType.PERSON:
            self.db.commit_person(obj, self.trans)
        elif note_type == NoteType.EVENT:
            self.db.commit_event(obj, self.trans)
        elif note_type == NoteType.FAMILY:
            self.db.commit_family(obj, self.trans)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2007-2008  Brian G. Matherly
# Copyright (C) 2009       Douglas S. Blank
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
#

"""
Display a people who have a person's same surname or given name.
"""

from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
ngettext = glocale.translation.ngettext  # else "nearby" comments are ignored
from gramps.gen.simple import SimpleAccess, SimpleDoc
from gramps.gui.plug.quick import QuickTable
from gramps.gen.lib import Person
from gramps.gen.filters.rules import Rule
from gramps.gen.filters import GenericFilterFactory


class IncompleteSurname(Rule):
    """People with incomplete surnames"""

    name = _("People with incomplete surnames")
    description = _("Matches people with lastname missing")
    category = _("General filters")

    def apply(self, db, person):
        for name in [person.get_primary_name()] + person.get_alternate_names():
            if name.get_group_name() == "":
                return True
        return False


class SameSurname(Rule):
    """People with same surname"""

    labels = [_("Substring:")]
    name = _("People matching the <surname>")
    description = _("Matches people with same lastname")
    category = _("General filters")

    def apply(self, db, person):
        src = self.list[0].upper()
        for name in [person.get_primary_name()] + person.get_alternate_names():
            if name.get_surname() and name.get_surname().upper() == src.upper():
                return True
        return False


class SameGiven(Rule):
    """People with same given name"""

    labels = [_("Substring:")]
    name = _("People matching the <given>")
    description = _("Matches people with same given name")
    category = _("General filters")

    def apply(self, db, person):
        src = self.list[0].upper()
        for name in [person.get_primary_name()] + person.get_alternate_names():
            if name.first_name:
                anyNBSP = name.first_name.split("\u00A0")
                if len(anyNBSP) > 1:  # there was an NBSP, a non-breaking space
                    first_two = anyNBSP[0] + "\u00A0" + anyNBSP[1].split()[0]
                    if first_two.upper() == src:
                        return True
                    else:
                        name.first_name = " ".join(anyNBSP[1].split()[1:])
                if " " in name.first_name.strip():
                    for name in name.first_name.upper().strip().split():
                        if name == src.upper():
                            return True
                elif name.first_name.upper() == src.upper():
                    return True
        return False


class IncompleteGiven(Rule):
    """People with incomplete given names"""

    name = _("People with incomplete given names")
    description = _("Matches people with firstname missing")
    category = _("General filters")

    def apply(self, db, person):
        for name in [person.get_primary_name()] + person.get_alternate_names():
            if name.get_first_name() == "":
                return True
        return False


def run(database, document, person):
    """
    Loops through the families that the person is a child in, and displays
    the information about the other children.
    """
    # setup the simple access functions
    sdb = SimpleAccess(database)
    sdoc = SimpleDoc(document)
    stab = QuickTable(sdb)
    if isinstance(person, Person):
        surname = sdb.surname(person)
        rsurname = person.get_primary_name().get_group_name()
    else:
        surname = person
        rsurname = person
    # display the title
    sdoc.title(_("People sharing the surname '%s'") % surname)
    sdoc.paragraph("")
    stab.columns(_("Person"), _("Birth Date"), _("Name type"))
    filter = GenericFilterFactory("Person")()
    if rsurname != "":
        rule = SameSurname([rsurname])
    else:
        rule = IncompleteSurname([])
    filter.add_rule(rule)
    people = filter.apply(database, database.iter_person_handles())

    matches = 0
    for person_handle in people:
        person = database.get_person_from_handle(person_handle)
        stab.row(
            person,
            sdb.birth_or_fallback(person),
            str(person.get_primary_name().get_type()),
        )
        matches += 1

    document.has_data = matches > 0
    sdoc.paragraph(
        # Translators: leave all/any {...} untranslated
        ngettext(
            "There is {number_of} person " "with a matching name, or alternate name.\n",
            "There are {number_of} people "
            "with a matching name, or alternate name.\n",
            matches,
        ).format(number_of=matches)
    )
    stab.write(sdoc)


def run_given(database, document, person):
    """
    Loops through the families that the person is a child in, and displays
    the information about the other children.
    """
    # setup the simple access functions
    sdb = SimpleAccess(database)
    sdoc = SimpleDoc(document)
    stab = QuickTable(sdb)
    if isinstance(person, Person):
        rgivenname = person.get_primary_name().get_first_name()
    else:
        rgivenname = person
    if " " in rgivenname.strip():
        rgivenname, second = rgivenname.strip().split(" ", 1)
    # display the title
    sdoc.title(_("People with the given name '%s'") % rgivenname)
    sdoc.paragraph("")
    stab.columns(_("Person"), _("Birth Date"), _("Name type"))
    filter = GenericFilterFactory("Person")()
    if rgivenname != "":
        rule = SameGiven([rgivenname])
    else:
        rule = IncompleteGiven([])
    filter.add_rule(rule)
    people = filter.apply(database, database.iter_person_handles())

    matches = 0
    for person_handle in people:
        person = database.get_person_from_handle(person_handle)
        stab.row(
            person,
            sdb.birth_or_fallback(person),
            str(person.get_primary_name().get_type()),
        )
        matches += 1

    document.has_data = matches > 0
    sdoc.paragraph(
        # Translators: leave all/any {...} untranslated
        ngettext(
            "There is {number_of} person " "with a matching name, or alternate name.\n",
            "There are {number_of} people "
            "with a matching name, or alternate name.\n",
            matches,
        ).format(number_of=matches)
    )
    stab.write(sdoc)



# -*- coding: utf-8 -*-
#!/usr/bin/env python
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2007       Johan Gonqvist <johan.gronqvist@gmail.com>
# Copyright (C) 2007-2009  Gary Burton <gary.burton@zen.co.uk>
# Copyright (C) 2007-2009  Stephane Charette <stephanecharette@gmail.com>
# Copyright (C) 2008-2009  Brian G. Matherly
# Copyright (C) 2008       Jason M. Simanek <jason@bohemianalps.com>
# Copyright (C) 2008-2011  Rob G. Healey <robhealey1@gmail.com>
# Copyright (C) 2010       Doug Blank <doug.blank@gmail.com>
# Copyright (C) 2010       Jakim Friant
# Copyright (C) 2010-      Serge Noiraud
# Copyright (C) 2011       Tim G L Lyons
# Copyright (C) 2013       Benny Malengier
# Copyright (C) 2016       Allen Crider
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Narrative Web Page generator.

Classe:
    RepositoryPage - Repository index page and individual Repository pages
"""
# ------------------------------------------------
# python modules
# ------------------------------------------------
from collections import defaultdict
from decimal import getcontext
import logging

# ------------------------------------------------
# Gramps module
# ------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale
from gramps.gen.lib import Repository
from gramps.plugins.lib.libhtml import Html

# ------------------------------------------------
# specific narrative web import
# ------------------------------------------------
from gramps.plugins.webreport.basepage import BasePage
from gramps.plugins.webreport.common import FULLCLEAR, html_escape

_ = glocale.translation.sgettext
LOG = logging.getLogger(".NarrativeWeb")
getcontext().prec = 8


#################################################
#
#    creates the Repository List Page and Repository Pages
#
#################################################
class RepositoryPages(BasePage):
    """
    This class is responsible for displaying information about the 'Repository'
    database objects. It displays this information under the 'Individuals'
    tab. It is told by the 'add_instances' call which 'Repository's to display,
    and remembers the list of persons. A single call to 'display_pages'
    displays both the Individual List (Index) page and all the Individual
    pages.

    The base class 'BasePage' is initialised once for each page that is
    displayed.
    """

    def __init__(self, report, the_lang, the_title):
        """
        @param: report    -- The instance of the main report class
                             for this report
        @param: the_lang  -- The lang to process
        @param: the_title -- The title page related to the language
        """
        BasePage.__init__(self, report, the_lang, the_title)
        self.repos_dict = defaultdict(set)

    def display_pages(self, the_lang, the_title):
        """
        Generate and output the pages under the Repository tab, namely the
        repository index and the individual repository pages.

        @param: the_lang  -- The lang to process
        @param: the_title -- The title page related to the language
        """
        LOG.debug("obj_dict[Person]")
        for item in self.report.obj_dict[Repository].items():
            LOG.debug("    %s", str(item))

        # set progress bar pass for Repositories
        message = _("Creating repository pages")
        progress_title = self.report.pgrs_title(the_lang)
        with self.r_user.progress(
            progress_title, message, len(self.report.obj_dict[Repository]) + 1
        ) as step:
            # Sort the repositories
            repos_dict = {}
            for repo_handle in self.report.obj_dict[Repository]:
                repository = self.r_db.get_repository_from_handle(repo_handle)
                key = repository.get_name() + str(repository.get_gramps_id())
                repos_dict[key] = (repository, repo_handle)

            keys = sorted(repos_dict, key=self.rlocale.sort_key)

            # RepositoryListPage Class
            self.repositorylistpage(self.report, the_lang, the_title, repos_dict, keys)

            idx = 1
            for dummy_index, key in enumerate(keys):
                (repo, handle) = repos_dict[key]
                step()
                idx += 1
                self.repositorypage(self.report, the_lang, the_title, repo, handle)

    def repositorylistpage(self, report, the_lang, the_title, repos_dict, keys):
        """
        Create Index for repositories

        @param: report     -- The instance of the main report class
                              for this report
        @param: the_lang   -- The lang to process
        @param: the_title  -- The title page related to the language
        @param: repos_dict -- The dictionary for all repositories
        @param: keys       -- The keys used to access repositories
        """
        BasePage.__init__(self, report, the_lang, the_title)
        # inc_repos = self.report.options["inc_repository"]

        output_file, sio = self.report.create_file("repositories")
        result = self.write_header(_("Repositories"))
        repolistpage, dummy_head, dummy_body, outerwrapper = result

        ldatec = 0
        # begin RepositoryList division
        with Html("div", class_="content", id="RepositoryList") as repositorylist:
            outerwrapper += repositorylist

            msg = self._(
                "This page contains an index of "
                "all the repositories in the "
                "database, sorted by their title. "
                "Clicking on a repositories&#8217;s title "
                "will take you to that repositories&#8217;s page."
            )
            repositorylist += Html("p", msg, id="description")

            # begin repositories table and table head
            with Html(
                "table", class_="infolist primobjlist repolist " + self.dir
            ) as table:
                repositorylist += table

                thead = Html("thead")
                table += thead

                trow = Html("tr") + (
                    Html("th", "&nbsp;", class_="ColumnRowLabel", inline=True),
                    Html("th", self._("Type"), class_="ColumnType", inline=True),
                    Html(
                        "th",
                        self._("Name", "Repository"),
                        class_="ColumnName",
                        inline=True,
                    ),
                )
                thead += trow

                # begin table body
                tbody = Html("tbody")
                table += tbody

                for index, key in enumerate(keys):
                    (repo, handle) = repos_dict[key]

                    trow = Html("tr")
                    tbody += trow

                    # index number
                    trow += Html("td", index + 1, class_="ColumnRowLabel", inline=True)

                    # repository type
                    rtype = self._(repo.type.xml_str())
                    trow += Html("td", rtype, class_="ColumnType", inline=True)

                    # repository name and hyperlink
                    if repo.get_name():
                        trow += Html(
                            "td",
                            self.repository_link(
                                handle,
                                repo.get_name(),
                                repo.get_gramps_id(),
                                self.uplink,
                            ),
                            class_="ColumnName",
                        )
                        ldatec = repo.get_change_time()
                    else:
                        trow += Html("td", "[ untitled ]", class_="ColumnName")

        # add clearline for proper styling
        # add footer section
        footer = self.write_footer(ldatec)
        outerwrapper += (FULLCLEAR, footer)

        # send page out for processing
        # and close the file
        self.xhtml_writer(repolistpage, output_file, sio, ldatec)

    def repositorypage(self, report, the_lang, the_title, repo, handle):
        """
        Create one page for one repository.

        @param: report    -- The instance of the main report class
                             for this report
        @param: the_lang  -- The lang to process
        @param: the_title -- The title page related to the language
        @param: repo      -- the repository to use
        @param: handle    -- the handle to use
        """
        gid = repo.get_gramps_id()
        BasePage.__init__(self, report, the_lang, the_title, gid)
        ldatec = repo.get_change_time()

        output_file, sio = self.report.create_file(handle, "repo")
        self.uplink = True
        result = self.write_header(_("Repositories"))
        repositorypage, dummy_head, dummy_body, outerwrapper = result

        # begin RepositoryDetail division and page title
        with Html("div", class_="content", id="RepositoryDetail") as repositorydetail:
            outerwrapper += repositorydetail

            # repository name
            repositorydetail += Html("h3", html_escape(repo.name), inline=True)

            # begin repository table
            with Html("table", class_="infolist repolist " + self.dir) as table:
                repositorydetail += table

                tbody = Html("tbody")
                table += tbody

                if not self.noid and gid:
                    trow = Html("tr") + (
                        Html(
                            "td",
                            self._("Gramps ID"),
                            class_="ColumnAttribute",
                            inline=True,
                        ),
                        Html("td", gid, class_="ColumnValue", inline=True),
                    )
                    tbody += trow

                trow = Html("tr") + (
                    Html("td", self._("Type"), class_="ColumnAttribute", inline=True),
                    Html(
                        "td",
                        self._(repo.get_type().xml_str()),
                        class_="ColumnValue",
                        inline=True,
                    ),
                )
                tbody += trow

            # Tags
            tags = self.show_tags(repo)
            if tags and self.report.inc_tags:
                trow = Html("tr") + (
                    Html("td", self._("Tags"), class_="ColumnAttribute", inline=True),
                    Html("td", tags, class_="ColumnValue", inline=True),
                )
                tbody += trow

            # repository: address(es)...
            # repository addresses do NOT have Sources
            repo_address = self.display_addr_list(repo.get_address_list(), False)
            if repo_address is not None:
                repositorydetail += repo_address

            # repository: urllist
            urllist = self.display_url_list(repo.get_url_list())
            if urllist is not None:
                repositorydetail += urllist

            # reposity: notelist
            notelist = self.display_note_list(repo.get_note_list(), Repository)
            if notelist is not None:
                repositorydetail += notelist

            # display Repository Referenced Sources...
            ref_list = self.display_bkref_list(Repository, repo.get_handle())
            if ref_list is not None:
                repositorydetail += ref_list

        # add clearline for proper styling
        # add footer section
        footer = self.write_footer(ldatec)
        outerwrapper += (FULLCLEAR, footer)

        # send page out for processing
        # and close the file
        self.xhtml_writer(repositorypage, output_file, sio, ldatec)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2001       Jesper Zedlitz
# Copyright (C) 2004-2006  Donald Allingham
# Copyright (C) 2007       Johan Gonqvist <johan.gronqvist@gmail.com>
# Copyright (C) 2008,2012  Brian G. Matherly
# Copyright (C) 2010       Jakim Friant
# Copyright (C) 2013-2014  Paul Franklin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""Reports/Text Reports /Number of Ancestors Report"""

# ------------------------------------------------------------------------
#
# standard python modules
#
# ------------------------------------------------------------------------
import math

# ------------------------------------------------------------------------
#
# Gramps modules
#
# ------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
from gramps.gen.errors import ReportError
from gramps.gen.plug.menu import PersonOption
from gramps.gen.plug.docgen import (
    IndexMark,
    FontStyle,
    ParagraphStyle,
    FONT_SANS_SERIF,
    PARA_ALIGN_CENTER,
    INDEX_TYPE_TOC,
)
from gramps.gen.plug.report import Report
from gramps.gen.plug.report import utils
from gramps.gen.plug.report import MenuReportOptions
from gramps.gen.plug.report import stdoptions
from gramps.gen.display.name import displayer as _nd


# ------------------------------------------------------------------------
#
# NumberOfAncestorsReport
#
# ------------------------------------------------------------------------
class NumberOfAncestorsReport(Report):
    """
    This report counts all the ancestors of the specified person.
    """

    def __init__(self, database, options, user):
        """
        Create the NumberOfAncestorsReport object that produces the report.

        The arguments are:

        database        - the Gramps database instance
        options         - instance of the Options class for this report
        user            - a gen.user.User() instance

        Menu options:
        name_format   - Preferred format to display names
        incl_private  - Whether to include private data
        """
        Report.__init__(self, database, options, user)

        stdoptions.run_private_data_option(self, options.menu)
        self.__db = self.database

        pid = options.menu.get_option_by_name("pid").get_value()
        self.__person = self.__db.get_person_from_gramps_id(pid)
        if self.__person is None:
            raise ReportError(_("Person %s is not in the Database") % pid)

        self.set_locale(options.menu.get_option_by_name("trans").get_value())

        stdoptions.run_name_format_option(self, options.menu)

    def write_report(self):
        """
        The routine that actually creates the report.
        At this point, the document is opened and ready for writing.
        """
        thisgen = {}
        all_people = {}
        total_theoretical = 0
        thisgen[self.__person.get_handle()] = 1
        ngettext = self._locale.translation.ngettext  # to see "nearby" comments

        self.doc.start_paragraph("NOA-Title")
        name = self._name_display.display(self.__person)
        # feature request 2356: avoid genitive form
        title = self._("Number of Ancestors for %s") % name
        mark = IndexMark(title, INDEX_TYPE_TOC, 1)
        self.doc.write_text(title, mark)
        self.doc.end_paragraph()

        thisgensize = 1
        gen = 0
        while thisgensize > 0:
            thisgensize = 0
            if thisgen != {}:
                thisgensize = len(thisgen)
                gen += 1
                theoretical = math.pow(2, (gen - 1))
                total_theoretical += theoretical
                percent = "(%s%%)" % self._locale.format_string(
                    "%3.2f", ((sum(thisgen.values()) / theoretical) * 100)
                )

                # TC # English return something like:
                # Generation 3 has 2 individuals. (50.00%)
                # Translators: leave all/any {...} untranslated
                text = ngettext(
                    "Generation {number} has {count} individual. {percent}",
                    "Generation {number} has {count} individuals. {percent}",
                    thisgensize,
                ).format(number=gen, count=thisgensize, percent=percent)

                self.doc.start_paragraph("NOA-Normal")
                self.doc.write_text(text)
                self.doc.end_paragraph()

            temp = thisgen
            thisgen = {}
            for person_handle, person_data in temp.items():
                person = self.__db.get_person_from_handle(person_handle)
                family_handle = person.get_main_parents_family_handle()
                if family_handle:
                    family = self.__db.get_family_from_handle(family_handle)
                    father_handle = family.get_father_handle()
                    mother_handle = family.get_mother_handle()
                    if father_handle:
                        thisgen[father_handle] = (
                            thisgen.get(father_handle, 0) + person_data
                        )
                        all_people[father_handle] = (
                            all_people.get(father_handle, 0) + person_data
                        )
                    if mother_handle:
                        thisgen[mother_handle] = (
                            thisgen.get(mother_handle, 0) + person_data
                        )
                        all_people[mother_handle] = (
                            all_people.get(mother_handle, 0) + person_data
                        )

        if total_theoretical != 1:
            percent = "(%3.2f%%)" % (
                (sum(all_people.values()) / (total_theoretical - 1)) * 100
            )
        else:
            percent = 0

        # TC # English return something like:
        # Total ancestors in generations 2 to 3 is 4. (66.67%)
        text = self._(
            "Total ancestors in generations %(second_generation)d "
            "to %(last_generation)d is %(count)d. %(percent)s"
        ) % {
            "second_generation": 2,
            "last_generation": gen,
            "count": len(all_people),
            "percent": percent,
        }

        self.doc.start_paragraph("NOA-Normal")
        self.doc.write_text(text)
        self.doc.end_paragraph()


# ------------------------------------------------------------------------
#
# NumberOfAncestorsOptions
#
# ------------------------------------------------------------------------
class NumberOfAncestorsOptions(MenuReportOptions):
    """
    Defines options for the NumberOfAncestorsReport.
    """

    def __init__(self, name, dbase):
        self.__db = dbase
        self.__pid = None
        MenuReportOptions.__init__(self, name, dbase)

    def get_subject(self):
        """Return a string that describes the subject of the report."""
        gid = self.__pid.get_value()
        person = self.__db.get_person_from_gramps_id(gid)
        return _nd.display(person)

    def add_menu_options(self, menu):
        """
        Add options to the menu for the Number of Ancestors report.
        """
        category_name = _("Report Options")

        self.__pid = PersonOption(_("Center Person"))
        self.__pid.set_help(_("The center person for the report"))
        menu.add_option(category_name, "pid", self.__pid)

        stdoptions.add_name_format_option(menu, category_name)

        stdoptions.add_private_data_option(menu, category_name)

        stdoptions.add_localization_option(menu, category_name)

    def make_default_style(self, default_style):
        """Make the default output style for the Number of Ancestors Report."""
        font = FontStyle()
        font.set_size(16)
        font.set_type_face(FONT_SANS_SERIF)
        font.set_bold(1)
        para = ParagraphStyle()
        para.set_header_level(1)
        para.set_bottom_border(1)
        para.set_bottom_margin(utils.pt2cm(8))
        para.set_font(font)
        para.set_alignment(PARA_ALIGN_CENTER)
        para.set_description(_("The style used for the title."))
        default_style.add_paragraph_style("NOA-Title", para)

        font = FontStyle()
        font.set_size(12)
        para = ParagraphStyle()
        para.set_font(font)
        para.set_description(_("The basic style used for the text display."))
        default_style.add_paragraph_style("NOA-Normal", para)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# gen.filters.rules/_ChangedSinceBase.py

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
import re
import time

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from . import Rule
from ...errors import FilterError
from ...const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext


# -------------------------------------------------------------------------
#
# ChangedSince
#
# -------------------------------------------------------------------------
class ChangedSinceBase(Rule):
    """
    Rule that checks for primary objects changed since a specific time.
    """

    labels = ["Changed after:", "but before:"]
    name = "Objects changed after <date time>"
    description = (
        "Matches object records changed after a specified "
        "date/time (yyyy-mm-dd hh:mm:ss) or in range, if a second "
        "date/time is given."
    )
    category = _("General filters")

    def add_time(self, date):
        if re.search(r"\d.*\s+\d{1,2}:\d{2}:\d{2}", date):
            return date
        elif re.search(r"\d.*\s+\d{1,2}:\d{2}", date):
            return date + ":00"
        elif re.search(r"\d.*\s+\d{1,2}", date):
            return date + ":00:00"
        elif re.search(r"\d{4}-\d{1,2}-\d{1,2}", date):
            return date + " 00:00:00"
        elif re.search(r"\d{4}-\d{1,2}", date):
            return date + "-01 00:00:00"
        elif re.search(r"\d{4}", date):
            return date + "-01-01 00:00:00"
        else:
            return date

    def time_str_to_sec(self, time_str):
        time_sec = None
        iso_date_time = self.add_time(time_str)
        try:
            time_tup = time.strptime(iso_date_time, "%Y-%m-%d %H:%M:%S")
            time_sec = time.mktime(time_tup)
        except ValueError:
            raise FilterError(
                _("Wrong format of date-time"),
                _(
                    "Only date-times in the iso format of yyyy-mm-dd "
                    "hh:mm:ss, where the time part is optional, are "
                    "accepted. %s does not satisfy."
                )
                % iso_date_time,
            )
        return time_sec

    def prepare(self, db, user):
        self.since = None
        self.before = None
        if self.list[0]:
            self.since = self.time_str_to_sec(self.list[0])
        if self.list[1]:
            self.before = self.time_str_to_sec(self.list[1])

    def apply(self, db, obj):
        obj_time = obj.get_change_time()
        if self.since:
            if obj_time < self.since:
                return False
            if self.before:
                return obj_time < self.before
            return True
        if self.before:
            return obj_time < self.before
        return False



# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2007-2009  Douglas S. Blank <doug.blank@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

# ------------------------------------------------------------------------
#
# Python modules
#
# ------------------------------------------------------------------------
from collections import defaultdict

# ------------------------------------------------------------------------
#
# Gramps modules
#
# ------------------------------------------------------------------------
from gramps.gen.plug import Gramplet
from gramps.gen.config import config
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext

# ------------------------------------------------------------------------
#
# Constants
#
# ------------------------------------------------------------------------

_YIELD_INTERVAL = 350


# ------------------------------------------------------------------------
#
# Local functions
#
# ------------------------------------------------------------------------
def make_tag_size(n, counts, mins=8, maxs=20):
    # return font sizes mins to maxs
    diff = maxs - mins
    # based on counts (biggest to smallest)
    if len(counts) > 1:
        position = diff - (diff * (float(counts.index(n)) / (len(counts) - 1)))
    else:
        position = 0
    return int(position) + mins


# ------------------------------------------------------------------------
#
# SurnameCloudGramplet class
#
# ------------------------------------------------------------------------
class SurnameCloudGramplet(Gramplet):
    def init(self):
        self.set_tooltip(_("Double-click surname for details"))
        self.top_size = 150  # will be overwritten in load
        self.min_font = 8
        self.max_font = 20
        self.set_text(_("No Family Tree loaded."))

    def db_changed(self):
        self.connect(self.dbstate.db, "person-add", self.update)
        self.connect(self.dbstate.db, "person-delete", self.update)
        self.connect(self.dbstate.db, "person-update", self.update)
        self.connect(self.dbstate.db, "person-rebuild", self.update)
        self.connect(self.dbstate.db, "family-rebuild", self.update)

    def on_load(self):
        if len(self.gui.data) == 3:
            self.top_size = int(self.gui.data[0])
            self.min_font = int(self.gui.data[1])
            self.max_font = int(self.gui.data[2])

    def save_update_options(self, widget=None):
        self.top_size = int(self.get_option(_("Number of surnames")).get_value())
        self.min_font = int(self.get_option(_("Min font size")).get_value())
        self.max_font = int(self.get_option(_("Max font size")).get_value())
        self.gui.data = [self.top_size, self.min_font, self.max_font]
        self.update()

    def main(self):
        self.set_text(_("Processing...") + "\n")
        yield True
        surnames = defaultdict(int)
        representative_handle = {}

        cnt = 0
        namelist = []
        for person in self.dbstate.db.iter_people():
            allnames = [person.get_primary_name()] + person.get_alternate_names()
            allnames = set([name.get_group_name().strip() for name in allnames])
            for surname in allnames:
                surnames[surname] += 1
                representative_handle[surname] = person.handle
            cnt += 1
            if not cnt % _YIELD_INTERVAL:
                yield True
            # Count unique surnames
            for name in [person.get_primary_name()] + person.get_alternate_names():
                if (
                    not name.get_surname().strip() in namelist
                    and not name.get_surname().strip() == ""
                ):
                    namelist.append(name.get_surname().strip())

        total_people = cnt
        surname_sort = []
        total = cnt = 0
        for surname in surnames:
            surname_sort.append((surnames[surname], surname))
            total += surnames[surname]
            cnt += 1
            if not cnt % _YIELD_INTERVAL:
                yield True

        surname_sort.sort(reverse=True)
        cloud_names = []
        cloud_values = []
        for count, surname in surname_sort:
            cloud_names.append((count, surname))
            cloud_values.append(count)

        cloud_names.sort(key=lambda k: k[1])
        line = 0
        ### All done!
        # Now, find out how many we can display without going over top_size:
        totals = defaultdict(int)
        for count, givensubname in cloud_names:  # givensubname_sort:
            totals[count] += 1
        sums = sorted(totals, reverse=True)
        total = 0
        include_greater_than = 0
        for s in sums:
            if total + totals[s] <= self.top_size:
                total += totals[s]
            else:
                include_greater_than = s
                break
        # now, limit counts to only include those that we can display:

        mins = self.min_font
        maxs = self.max_font
        # Ok, now we can show those counts > include_greater_than:
        good_counts = []
        for count, surname in cloud_names:  # surname_sort:
            if count > include_greater_than:
                good_counts.append(count)
        counts = list(set(good_counts))
        counts.sort(reverse=True)
        showing = 0
        self.set_text("")
        for count, surname in cloud_names:  # surname_sort:
            if count > include_greater_than:
                if len(surname) == 0:
                    text = config.get("preferences.no-surname-text")
                else:
                    text = surname
                size = make_tag_size(count, counts, mins=mins, maxs=maxs)
                self.link(
                    text,
                    "Surname",
                    representative_handle[surname],
                    size,
                    "%s, %d%% (%d)"
                    % (text, int((float(count) / total_people) * 100), count),
                )
                self.append_text(" ")
                showing += 1
        self.append_text(
            ("\n\n" + _("Total unique surnames") + ": %d\n") % len(namelist)
        )
        self.append_text((_("Total surnames showing") + ": %d\n") % showing)
        self.append_text((_("Total people") + ": %d") % total_people, "begin")

    def build_options(self):
        from gramps.gen.plug.menu import NumberOption

        self.top_size_option = NumberOption(
            _("Number of surnames"), self.top_size, 1, 150
        )
        self.add_option(self.top_size_option)
        self.min_option = NumberOption(_("Min font size"), self.min_font, 1, 50)
        self.add_option(self.min_option)
        self.max_option = NumberOption(_("Max font size"), self.max_font, 1, 50)
        self.add_option(self.max_option)

    def save_options(self):
        self.top_size = int(self.get_option(_("Number of surnames")).get_value())
        self.min_font = int(self.get_option(_("Min font size")).get_value())
        self.max_font = int(self.get_option(_("Max font size")).get_value())



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2013       Doug Blank <doug.blank@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Researcher information for Gramps.
"""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .locationbase import LocationBase


# -------------------------------------------------------------------------
#
#
#
# -------------------------------------------------------------------------
class Researcher(LocationBase):
    """Contains the information about the owner of the database."""

    def __init__(self, source=None):
        """
        Initialize the Researcher object, copying from the source if provided.
        """

        LocationBase.__init__(self, source)
        if source:
            self.name = source.name
            self.addr = source.addr
            self.email = source.email
        else:
            self.name = ""
            self.addr = ""
            self.email = ""

    def serialize(self):
        """
        Convert the object to a serialized tuple of data.
        """
        return (LocationBase.serialize(self), self.name, self.addr, self.email)

    def unserialize(self, data):
        """
        Convert a serialized tuple of data to an object.
        """
        (location, self.name, self.addr, self.email) = data
        LocationBase.unserialize(self, location)

        return self

    def set_name(self, data):
        """Set the database owner's name."""
        self.name = data

    def get_name(self):
        """Return the database owner's name."""
        return self.name

    def set_address(self, data):
        """Set the database owner's address."""
        self.addr = data

    def get_address(self):
        """Return the database owner's address."""
        return self.addr

    def set_email(self, data):
        """Set the database owner's email."""
        self.email = data

    def get_email(self):
        """Return the database owner's email."""
        return self.email

    def set_from(self, other_researcher):
        """Set all attributes from another instance."""
        self.street = other_researcher.street
        self.locality = other_researcher.locality
        self.city = other_researcher.city
        self.county = other_researcher.county
        self.state = other_researcher.state
        self.country = other_researcher.country
        self.postal = other_researcher.postal
        self.phone = other_researcher.phone

        self.name = other_researcher.name
        self.addr = other_researcher.addr
        self.email = other_researcher.email

    def get(self):
        return [
            getattr(self, value)
            for value in [
                "name",
                "addr",
                "locality",
                "city",
                "state",
                "country",
                "postal",
                "phone",
                "email",
            ]
        ]

    def is_empty(self):
        for attr in [
            "name",
            "addr",
            "locality",
            "city",
            "state",
            "country",
            "postal",
            "phone",
            "email",
        ]:
            if getattr(self, attr) != "":
                return False
        return True



# -*- coding: utf-8 -*-
#!/usr/bin/env python
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2007       Johan Gonqvist <johan.gronqvist@gmail.com>
# Copyright (C) 2007-2009  Gary Burton <gary.burton@zen.co.uk>
# Copyright (C) 2007-2009  Stephane Charette <stephanecharette@gmail.com>
# Copyright (C) 2008-2009  Brian G. Matherly
# Copyright (C) 2008       Jason M. Simanek <jason@bohemianalps.com>
# Copyright (C) 2008-2011  Rob G. Healey <robhealey1@gmail.com>
# Copyright (C) 2010       Doug Blank <doug.blank@gmail.com>
# Copyright (C) 2010       Jakim Friant
# Copyright (C) 2010-      Serge Noiraud
# Copyright (C) 2011       Tim G L Lyons
# Copyright (C) 2013       Benny Malengier
# Copyright (C) 2016       Allen Crider
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Narrative Web Page generator.

Classe:
    DownloadPage
"""
# ------------------------------------------------
# python modules
# ------------------------------------------------
import os
import datetime
from decimal import getcontext
import logging

# ------------------------------------------------
# Gramps module
# ------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale
from gramps.plugins.lib.libhtml import Html

# ------------------------------------------------
# specific narrative web import
# ------------------------------------------------
from gramps.plugins.webreport.basepage import BasePage
from gramps.plugins.webreport.common import FULLCLEAR, html_escape
from gramps.gen.utils.file import create_checksum

_ = glocale.translation.sgettext
LOG = logging.getLogger(".NarrativeWeb")
getcontext().prec = 8


class DownloadPage(BasePage):
    """
    This class is responsible for displaying information about the Download page
    """

    def __init__(self, report, the_lang, the_title):
        """
        @param: report    -- The instance of the main report class
                             for this report
        @param: the_lang  -- The lang to process
        @param: the_title -- The title page related to the language
        """
        BasePage.__init__(self, report, the_lang, the_title)

        # do NOT include a Download Page
        if not self.report.inc_download:
            return

        # menu options for class
        # download and description #n ( 1 <= n < 5 )

        dlfname = self.report.dl_fname
        dldescr = self.report.dl_descr

        # if no filenames at all, return???
        if dlfname:
            output_file, sio = self.report.create_file("download")
            result = self.write_header(self._("Download"))
            downloadpage, dummy_head, dummy_body, outerwrapper = result

            # begin download page and table
            with Html("div", class_="content", id="Download") as download:
                outerwrapper += download

                msg = self._(
                    "This page is for the user/ creator "
                    "of this Family Tree/ Narrative website "
                    "to share a couple of files with you "
                    "regarding their family.  If there are "
                    "any files listed "
                    "below, clicking on them will allow you "
                    "to download them. The "
                    "download page and files have the same "
                    "copyright as the remainder "
                    "of these web pages."
                )
                download += Html("p", msg, id="description")

                # begin download table and table head
                with Html("table", class_="infolist download") as table:
                    download += table

                    thead = Html("thead")
                    table += thead

                    trow = Html("tr")
                    thead += trow

                    trow.extend(
                        Html("th", label, class_="Column" + colclass, inline=True)
                        for (label, colclass) in [
                            (self._("File Name"), "Filename"),
                            (self._("Description"), "Description"),
                            (self._("Last Modified"), "Modified"),
                            (self._("MD5"), "Md5"),
                        ]
                    )
                    # table body
                    tbody = Html("tbody")
                    table += tbody
                    dwnld = 0

                    for fnamex in dlfname:
                        # if fnamex is not None, do we have a file to download?
                        if fnamex:
                            fname = os.path.basename(dlfname[fnamex])
                            # if fname is not None, show it
                            if fname:
                                dwnld += 1
                                trow = Html("tr", id="Row01")
                                tbody += trow
                                fname_lnk = "../" + fname if the_lang else fname
                                dldescrx = dldescr[fnamex]
                                tcell = Html("td", class_="ColumnFilename") + (
                                    Html(
                                        "a",
                                        fname,
                                        href=fname_lnk,
                                        title=html_escape(dldescrx),
                                    )
                                )
                                trow += tcell

                                dldescr1 = dldescrx or "&nbsp;"
                                trow += Html(
                                    "td",
                                    dldescr1,
                                    inline=True,
                                    class_="ColumnDescription",
                                )

                                tcell = Html("td", class_="ColumnModified", inline=True)
                                trow += tcell
                                if os.path.exists(dlfname[fnamex]):
                                    md5 = create_checksum(dlfname[fnamex])
                                    trow += Html(
                                        "td", md5, class_="ColumnMd5", inline=True
                                    )
                                    modified = os.stat(dlfname[fnamex]).st_mtime
                                    last_mod = datetime.datetime.fromtimestamp(modified)
                                    tcell += last_mod
                                    # copy the file only once
                                    if not os.path.exists(fname):
                                        self.report.copy_file(dlfname[fnamex], fname)
                                else:
                                    tcell += self._("Cannot open file")

                    if not dwnld:
                        # We have several files to download
                        # but all file names are empty
                        dldescrx = self._("No file to download")
                        trow = Html("tr", id="Row01")
                        tbody += trow
                        tcell = Html("td", class_="ColumnFilename", colspan=3) + Html(
                            "h4", dldescrx
                        )
                        trow += tcell
            # clear line for proper styling
            # create footer section
            footer = self.write_footer(None)
            outerwrapper += (FULLCLEAR, footer)

            # send page out for processing
            # and close the file
            self.xhtml_writer(downloadpage, output_file, sio, 0)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

import logging


class RotateHandler(logging.Handler):
    """
    A handler class which buffers logging records in memory. A rotating
    buffer is used so that the last X records are always available.
    """

    def __init__(self, capacity):
        """
        Initialize the handler with the buffer size.
        """
        logging.Handler.__init__(self)

        self.set_capacity(capacity)

    def emit(self, record):
        """
        Add the record to the rotating buffer.

        """
        self._buffer[self._index] = record
        self._index = (self._index + 1) % self._capacity

    def get_buffer(self):
        """
        Return the buffer with the records in the correct order.
        """

        return [
            _f for _f in self._buffer[self._index :] + self._buffer[: self._index] if _f
        ]

    def get_formatted_log(self, remove_tail_duplicate=None):
        """
        Return the log buffer after it has been formatted.

        Returns a list of strings.
        """

        return [
            self.format(record)
            for record in self._buffer[self._index :] + self._buffer[: self._index]
            if record is not None and record != remove_tail_duplicate
        ]

    def set_capacity(self, capacity):
        """
        Set the number of log records that will be stored.
        """
        self._capacity = capacity
        self._index = 0
        self._buffer = self._capacity * [None]



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Base Reference class for Gramps.
"""

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from abc import ABCMeta, abstractmethod


# -------------------------------------------------------------------------
#
# RefBase class
#
# -------------------------------------------------------------------------
class RefBase(metaclass=ABCMeta):
    """
    Base reference class to manage references to other objects.

    Any *Ref* classes should derive from this class.
    """

    def __init__(self, source=None):
        if source:
            self.ref = source.ref
        else:
            self.ref = None

    def serialize(self):
        """
        Convert the object to a serialized tuple of data.
        """
        return self.ref

    def unserialize(self, data):
        """
        Convert a serialized tuple of data to an object.
        """
        self.ref = data
        return self

    @abstractmethod
    def get_referenced_handles(self):
        """
        Returns the list of (classname, handle) tuples for all directly
        referenced primary objects.

        :returns: Returns the list of (classname, handle) tuples for referenced
                  objects.
        :rtype: list
        """

    def set_reference_handle(self, handle):
        """
        Set the reference handle.

        :param handle: The reference handle.
        :type handle: str
        """
        self.ref = handle

    def get_reference_handle(self):
        """
        Return the reference handle.

        :returns: The reference handle.
        :rtype: str
        """
        return self.ref



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2009  Florian Heinle
# Copyright (C) 2010  Doug Blank <doug.blank@gmail.com>
# Copyright (C) 2010  Benny Malengier
# Copyright (C) 2014  Vassilii Khachaturov
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
gtk textbuffer with undo functionality
"""

__all__ = ["UndoableStyledBuffer"]

from contextlib import contextmanager

from gi.repository import Gtk

from gramps.gen.lib.styledtext import StyledText
from .undoablebuffer import UndoableInsert, UndoableDelete
from .styledtextbuffer import StyledTextBuffer


class UndoableInsertStyled(UndoableInsert):
    """something that has been inserted into our styledtextbuffer"""

    def __init__(self, text_iter, text, length, text_buffer):
        # we obtain the buffer before the text has been inserted
        UndoableInsert.__init__(self, text_iter, text, length, text_buffer)
        self.tags = text_buffer.get_text(
            text_buffer.get_start_iter(), text_buffer.get_end_iter(), True
        ).get_tags()
        self.tagsafter = None


class UndoableDeleteStyled(UndoableDelete):
    def __init__(self, text_buffer, start_iter, end_iter):
        # we obtain the buffer before the text has been deleted
        UndoableDelete.__init__(self, text_buffer, start_iter, end_iter)
        self.tags = text_buffer.get_text(
            text_buffer.get_start_iter(), text_buffer.get_end_iter(), True
        ).get_tags()


class UndoableApplyStyle:
    """a style has been applied to our textbuffer"""

    def __init__(self, text_buffer, tag, start, end):
        self.offset = text_buffer.get_iter_at_mark(
            text_buffer.get_insert()
        ).get_offset()

        self.mergeable = False
        self.tags = text_buffer.get_text(
            text_buffer.get_start_iter(), text_buffer.get_end_iter(), True
        ).get_tags()
        #
        self.tags_after = None
        self.offset_after = None

    def set_after(self, tags, offset):
        self.tags_after = tags
        self.offset_after = offset


class UndoableStyledBuffer(StyledTextBuffer):
    """text buffer with added undo capabilities for styledtextbuffer

    designed as a drop-in replacement for gtksourceview,
    at least as far as undo is concerned"""

    insertclass = UndoableInsertStyled
    deleteclass = UndoableDeleteStyled

    def __init__(self):
        StyledTextBuffer.__init__(self)
        self.connect("apply-tag", self.on_tag_insert_undoable)
        self.connect_after("apply-tag", self.on_tag_afterinsert_undoable)

    @contextmanager
    def undo_disabled(self):
        """
        Assures that not_undoable_action is False during the context.

        Usage example (see gramps/gui/widgets/styledtexteditor.py)::

            with self.buffer.undo_disabled():
                ... # heavy stuff like spell checking
        """
        oldflag = self.not_undoable_action
        self.not_undoable_action = True
        try:
            yield
        except:
            raise
        finally:
            self.not_undoable_action = oldflag

    def on_tag_insert_undoable(self, buffer, tag, start, end):
        if tag.get_property("name") == "hyperlink":
            return
        if not self.undo_in_progress:
            self._empty_redo_stack()
        if self.not_undoable_action:
            return
        if end.get_offset() - start.get_offset() == 1:
            # only store this 1 character tag if in a different place
            if (
                self.undo_stack
                and isinstance(self.undo_stack[-1], UndoableInsertStyled)
                and self.undo_stack[-1].offset + self.undo_stack[-1].length
                == end.get_offset()
            ):
                return
        undo_action = UndoableApplyStyle(buffer, tag, start, end)
        self.undo_stack.append(undo_action)

    def on_tag_afterinsert_undoable(self, buffer, tag, start, end):
        if self.not_undoable_action:
            return
        if not self.undo_stack or not isinstance(
            self.undo_stack[-1], UndoableApplyStyle
        ):
            return
        self.undo_stack[-1].set_after(
            buffer.get_text(
                buffer.get_start_iter(), buffer.get_end_iter(), True
            ).get_tags(),
            buffer.get_iter_at_mark(buffer.get_insert()).get_offset(),
        )

    def _undo_insert(self, undo_action):
        start = self.get_iter_at_offset(undo_action.offset)
        stop = self.get_iter_at_offset(undo_action.offset + undo_action.length)
        self.delete(start, stop)
        # the text is correct again, now we create correct styled text
        s_text = StyledText(
            Gtk.TextBuffer.get_text(
                self, self.get_start_iter(), self.get_end_iter(), True
            ),
            undo_action.tags,
        )
        self.set_text(s_text)
        self.place_cursor(self.get_iter_at_offset(undo_action.offset))

    def _undo_delete(self, undo_action):
        start = self.get_iter_at_offset(undo_action.start)
        self.insert(start, undo_action.text)
        # the text is correct again, now we create correct styled text
        s_text = StyledText(
            Gtk.TextBuffer.get_text(
                self, self.get_start_iter(), self.get_end_iter(), True
            ),
            undo_action.tags,
        )
        self.set_text(s_text)
        if undo_action.delete_key_used:
            self.place_cursor(self.get_iter_at_offset(undo_action.start))
        else:
            self.place_cursor(self.get_iter_at_offset(undo_action.end))

    def _redo_insert(self, redo_action):
        s_text = StyledText(
            Gtk.TextBuffer.get_text(
                self, self.get_start_iter(), self.get_end_iter(), True
            ),
            redo_action.tags,
        )
        self.set_text(s_text)
        start = self.get_iter_at_offset(redo_action.offset)
        self.insert(start, redo_action.text)
        new_cursor_pos = self.get_iter_at_offset(
            redo_action.offset + redo_action.length
        )
        self.place_cursor(new_cursor_pos)

    def _redo_delete(self, redo_action):
        start = self.get_iter_at_offset(redo_action.start)
        stop = self.get_iter_at_offset(redo_action.end)
        self.delete(start, stop)
        # the text is correct again, now we create correct styled text
        # s_text = StyledText(Gtk.TextBuffer.get_text(self,
        #    self.get_start_iter(), self.get_end_iter(), True), redo_action.tags)
        # self.set_text(s_text)
        self.place_cursor(self.get_iter_at_offset(redo_action.start))

    def _handle_undo(self, undo_action):
        """undo of apply of style"""
        s_text = StyledText(
            Gtk.TextBuffer.get_text(
                self, self.get_start_iter(), self.get_end_iter(), True
            ),
            undo_action.tags,
        )
        self.set_text(s_text)
        self.place_cursor(self.get_iter_at_offset(undo_action.offset))

    def _handle_redo(self, redo_action):
        """redo of apply of style"""
        s_text = StyledText(
            Gtk.TextBuffer.get_text(
                self, self.get_start_iter(), self.get_end_iter(), True
            ),
            redo_action.tags_after,
        )
        self.set_text(s_text)
        self.place_cursor(self.get_iter_at_offset(redo_action.offset_after))



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2009 B. Malengier
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
General utility functions useful for the generic plugin system
"""

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
import sys
import os
import datetime
from io import StringIO, BytesIO
import json

# -------------------------------------------------------------------------
#
# set up logging
#
# -------------------------------------------------------------------------
import logging

LOG = logging.getLogger(".gen.plug")

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ._pluginreg import make_environment
from ..const import USER_PLUGINS
from ...version import VERSION_TUPLE
from . import BasePluginManager
from ..utils.configmanager import safe_eval
from ..config import config
from ..constfunc import mac
from ..const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext


# -------------------------------------------------------------------------
#
# Local utility functions for gen.plug
#
# -------------------------------------------------------------------------
def version_str_to_tup(sversion, positions):
    """
    Given a string version and positions count, returns a tuple of
    integers.

    >>> version_str_to_tup("1.02.9", 2)
    (1, 2)
    """
    try:
        tup = tuple(map(int, sversion.split(".")))
        tup += (0,) * (positions - len(tup))
    except:
        tup = (0,) * positions
    return tup[:positions]


class FakeRegistrar:
    """
    Fakes ``register`` and ``newplugin``.
    """

    def __init__(self):
        self.results = []

    @property
    def newplugin(self):
        class _newplugin:
            """
            Fake newplugin.
            """

            def __init__(inner_self):
                self.results.append({})

            def __setattr__(inner_self, attr, value):
                self.results[-1][attr] = value

        return _newplugin

    def register(self, ptype, **kwargs):
        """
        Fake registration. Side-effect sets results to kwargs.
        """
        retval = {"ptype": ptype}
        retval.update(kwargs)
        # Get the results back to calling function
        self.results.append(retval)


class Zipfile:
    """
    Class to duplicate the methods of tarfile.TarFile, for Python 2.5.
    """

    def __init__(self, buffer):
        import zipfile

        self.buffer = buffer
        self.zip_obj = zipfile.ZipFile(buffer)

    def extractall(self, path, members=None):
        """
        Extract all of the files in the zip into path.
        """
        names = self.zip_obj.namelist()
        for name in self.get_paths(names):
            fullname = os.path.join(path, name)
            if not os.path.exists(fullname):
                os.mkdir(fullname)
        for name in self.get_files(names):
            fullname = os.path.join(path, name)
            with open(fullname, "wb") as outfile:
                outfile.write(self.zip_obj.read(name))

    def extractfile(self, name):
        """
        Extract a name from the zip file.

        >>> Zipfile(buffer).extractfile("Dir/dile.py").read()
        <Contents>
        """

        class ExtractFile:
            def __init__(self, zip_obj, name):
                self.zip_obj = zip_obj
                self.name = name

            def read(self):
                data = self.zip_obj.read(self.name)
                del self.zip_obj
                return data

        return ExtractFile(self.zip_obj, name)

    def close(self):
        """
        Close the zip object.
        """
        self.zip_obj.close()

    def getnames(self):
        """
        Get the files and directories of the zipfile.
        """
        return self.zip_obj.namelist()

    def get_paths(self, items):
        """
        Get the directories from the items.
        """
        return (name for name in items if self.is_path(name) and not self.is_file(name))

    def get_files(self, items):
        """
        Get the files from the items.
        """
        return (name for name in items if self.is_file(name))

    def is_path(self, name):
        """
        Is the name a path?
        """
        return os.path.split(name)[0]

    def is_file(self, name):
        """
        Is the name a directory?
        """
        return os.path.split(name)[1]


def urlopen_maybe_no_check_cert(URL):
    """
    Similar to urllib.request.urlopen, but disables certificate
    verification on Mac.
    """
    context = None
    from urllib.request import urlopen

    if mac():
        from ssl import create_default_context, CERT_NONE

        context = create_default_context()
        context.check_hostname = False
        context.verify_mode = CERT_NONE
    timeout = 10  # seconds
    fp = None
    try:
        fp = urlopen(URL, timeout=timeout, context=context)
    except TypeError:
        fp = urlopen(URL, timeout=timeout)
    return fp


def get_addons(project, url):
    LOG.debug("Checking for updated addons...")
    if not url.startswith(("http://", "https://", "file://")):
        return []
    langs = glocale.get_language_list()
    langs.append("en")
    # now we have a list of languages to try:
    fp = None
    for lang in langs:
        URL = "%s/listings/addons-%s.json" % (url, lang)
        LOG.debug("   trying: %s" % URL)
        try:
            fp = urlopen_maybe_no_check_cert(URL)
        except:
            try:
                URL = "%s/listings/addons-%s.json" % (url, lang[:2])
                fp = urlopen_maybe_no_check_cert(URL)
            except Exception as err:  # some error
                LOG.warning(
                    "Failed to open addon metadata for {lang} {url}: {err}".format(
                        lang=lang, url=URL, err=err
                    )
                )
                fp = None
        if fp and (fp.getcode() == 200 or fp.file):
            break

    addon_list = []
    if fp and (fp.getcode() == 200 or fp.file):
        addon_list = json.load(fp)
        for plugin_dict in addon_list:
            if "a" not in plugin_dict:
                plugin_dict["a"] = 0
            if "s" not in plugin_dict:
                plugin_dict["s"] = 0
            if "h" not in plugin_dict:
                plugin_dict["h"] = ""
            plugin_dict["_p"] = project
            plugin_dict["_u"] = url
            id = plugin_dict["i"]
            pmgr = BasePluginManager.get_instance()
            plugin = pmgr.get_plugin(id)
            if plugin:
                plugin_dict["_v"] = plugin.version
    else:
        LOG.debug("Checking Addons Failed")
    LOG.debug("Done checking!")

    return addon_list


def get_all_addons():
    projects = config.get("behavior.addons-projects")
    all_addons = []
    for project, url, enabled in projects:
        if enabled:
            addons_list = get_addons(project, url)
            all_addons.extend(addons_list)
    return all_addons


def available_updates():
    whattypes = config.get("behavior.check-for-addon-update-types")
    addon_update_list = []
    for plugin_dict in get_all_addons():
        if "_v" in plugin_dict:
            LOG.debug(
                "Comparing %s > %s"
                % (
                    version_str_to_tup(plugin_dict["v"], 3),
                    version_str_to_tup(plugin_dict["_v"], 3),
                )
            )
            if version_str_to_tup(plugin_dict["v"], 3) > version_str_to_tup(
                plugin_dict["_v"], 3
            ):
                LOG.debug("   Downloading '%s'..." % plugin_dict["z"])
                if "update" in whattypes:
                    if not config.get(
                        "behavior.do-not-show-previously-seen-addon-updates"
                    ) or plugin_dict["i"] not in config.get(
                        "behavior.previously-seen-addon-updates"
                    ):
                        addon_update_list.append(
                            (
                                _("Updated"),
                                "%s/download/%s"
                                % (plugin_dict["_u"], plugin_dict["z"]),
                                plugin_dict,
                            )
                        )
            else:
                LOG.debug("   '%s' is ok" % plugin_dict["n"])
        else:
            LOG.debug("   '%s' is not installed" % plugin_dict["n"])
            if "new" in whattypes:
                if not config.get(
                    "behavior.do-not-show-previously-seen-addon-updates"
                ) or plugin_dict["i"] not in config.get(
                    "behavior.previously-seen-addon-updates"
                ):
                    addon_update_list.append(
                        (
                            _("New", "updates"),
                            "%s/download/%s" % (plugin_dict["_u"], plugin_dict["z"]),
                            plugin_dict,
                        )
                    )
    config.set(
        "behavior.last-check-for-addon-updates",
        datetime.date.today().strftime("%Y/%m/%d"),
    )

    return addon_update_list


def load_addon_file(path, callback=None):
    """
    Load an addon from a particular path (from URL or file system).
    """
    from urllib.request import urlopen
    import tarfile

    if (
        path.startswith("http://")
        or path.startswith("https://")
        or path.startswith("ftp://")
        or path.startswith("file://")
    ):
        try:
            fp = urlopen_maybe_no_check_cert(path)
        except:
            if callback:
                callback(_("Unable to open '%s'") % path)
            return False
    else:
        try:
            fp = open(path, "rb")
        except:
            if callback:
                callback(_("Unable to open '%s'") % path)
            return False
    try:
        content = fp.read()
        buffer = BytesIO(content)
    except:
        if callback:
            callback(_("Error in reading '%s'") % path)
        return False
    fp.close()
    # file_obj is either Zipfile or TarFile
    if path.endswith(".zip") or path.endswith(".ZIP"):
        file_obj = Zipfile(buffer)
    elif path.endswith(".tar.gz") or path.endswith(".tgz"):
        try:
            file_obj = tarfile.open(None, fileobj=buffer)
        except:
            if callback:
                callback(_("Error: cannot open '%s'") % path)
            return False
    else:
        if callback:
            callback(_("Error: unknown file type: '%s'") % path)
        return False
    # First, see what versions we have/are getting:
    good_gpr = set()
    for gpr_file in [name for name in file_obj.getnames() if name.endswith(".gpr.py")]:
        if callback:
            callback((_("Examining '%s'...") % gpr_file) + "\n")
        contents = file_obj.extractfile(gpr_file).read()
        registrar = FakeRegistrar()
        # Put a fake register and _ function in environment:
        env = make_environment(
            register=registrar.register,
            newplugin=registrar.newplugin,
            _=lambda text: text,
        )
        # evaluate the contents:
        try:
            exec(contents, env)
        except Exception as exp:
            if callback:
                msg = _("Error in '%s' file: cannot load.") % gpr_file
                callback("   " + msg + "\n" + str(exp))
            continue
        # There can be multiple addons per gpr file:
        for results in registrar.results:
            gramps_target_version = results.get("gramps_target_version", None)
            id = results.get("id", None)
            if gramps_target_version:
                vtup = version_str_to_tup(gramps_target_version, 2)
                # Is it for the right version of gramps?
                if vtup == VERSION_TUPLE[0:2]:
                    # If this version is not installed, or > installed, install it
                    good_gpr.add(gpr_file)
                    if callback:
                        callback(
                            "   "
                            + (_("'%s' is for this version of Gramps.") % id)
                            + "\n"
                        )
                else:
                    # If the plugin is for another version; inform and do nothing
                    if callback:
                        callback(
                            "   "
                            + (_("'%s' is NOT for this version of Gramps.") % id)
                            + "\n"
                        )
                        callback(
                            "   "
                            + (
                                _("It is for version %(v1)d.%(v2)d")
                                % {"v1": vtup[0], "v2": vtup[1]}
                                + "\n"
                            )
                        )
                    continue
            else:
                # another register function doesn't have gramps_target_version
                if gpr_file in good_gpr:
                    s.remove(gpr_file)
                if callback:
                    callback(
                        "   "
                        + (
                            _("Error: missing gramps_target_version in '%s'...")
                            % gpr_file
                        )
                        + "\n"
                    )
    registered_count = 0
    if len(good_gpr) > 0:
        # Now, install the ok ones
        try:
            file_obj.extractall(USER_PLUGINS)
        except OSError:
            if callback:
                callback("OSError installing '%s', skipped!" % path)
            file_obj.close()
            return False
        if callback:
            callback((_("Installing '%s'...") % path) + "\n")
        gpr_files = set(
            [os.path.split(os.path.join(USER_PLUGINS, name))[0] for name in good_gpr]
        )
        for gpr_file in gpr_files:
            if callback:
                callback("   " + (_("Registered '%s'") % gpr_file) + "\n")
            registered_count += 1
    file_obj.close()
    if registered_count:
        return True
    else:
        return False


# -------------------------------------------------------------------------
#
# OpenFileOrStdout class
#
# -------------------------------------------------------------------------
class OpenFileOrStdout:
    """Context manager to open a file or stdout for writing."""

    def __init__(self, filename, encoding=None, errors=None, newline=None):
        self.filename = filename
        self.filehandle = None
        self.encoding = encoding
        self.errors = errors
        self.newline = newline

    def __enter__(self):
        if self.filename == "-":
            self.filehandle = sys.stdout
        else:
            self.filehandle = open(
                self.filename,
                "w",
                encoding=self.encoding,
                errors=self.errors,
                newline=self.newline,
            )
        return self.filehandle

    def __exit__(self, exc_type, exc_value, traceback):
        if self.filehandle and self.filename != "-":
            self.filehandle.close()
        return False


# -------------------------------------------------------------------------
#
# OpenFileOrStdin class
#
# -------------------------------------------------------------------------
class OpenFileOrStdin:
    """Context manager to open a file or stdin for reading."""

    def __init__(self, filename, add_mode="", encoding=None):
        self.filename = filename
        self.mode = "r%s" % add_mode
        self.filehandle = None
        self.encoding = encoding

    def __enter__(self):
        if self.filename == "-":
            self.filehandle = sys.stdin
        elif self.encoding:
            self.filehandle = open(self.filename, self.mode, encoding=self.encoding)
        else:
            self.filehandle = open(self.filename, self.mode)
        return self.filehandle

    def __exit__(self, exc_type, exc_value, traceback):
        if self.filename != "-":
            self.filehandle.close()
        return False


# -------------------------------------------------------------------------
#
# get_cite function
#
# -------------------------------------------------------------------------
def get_cite():
    """
    Function that returns the active cite plugin.
    """
    plugman = BasePluginManager.get_instance()
    for pdata in plugman.get_reg_cite():
        if pdata.id != config.get("preferences.cite-plugin"):
            continue
        module = plugman.load_plugin(pdata)
        if not module:
            print("Error loading formatter '%s': skipping content" % pdata.name)
            continue
        cite = getattr(module, "Formatter")()
        return cite



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2007 Zsolt Foldvari
# Copyright (C) 2008 Brian G. Matherly
# Copyright (C) 2013 Vassilii Khachaturov
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""PS/PDF output generator based on Cairo.
"""

# ------------------------------------------------------------------------
#
# Python modules
#
# ------------------------------------------------------------------------
import logging

# -------------------------------------------------------------------------
#
# GTK modules
#
# -------------------------------------------------------------------------
import gi

gi.require_version("Pango", "1.0")
gi.require_version("PangoCairo", "1.0")
from gi.repository import Pango, PangoCairo
import cairo

# ------------------------------------------------------------------------
#
# Gramps modules
#
# ------------------------------------------------------------------------
from gramps.gen.constfunc import is_quartz
import gramps.plugins.lib.libcairodoc as libcairodoc
from gramps.gen.plug.docgen import INDEX_TYPE_ALP, INDEX_TYPE_TOC
from gramps.gen.errors import ReportError
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# ------------------------------------------------------------------------
#
# Set up logging
#
# ------------------------------------------------------------------------
LOG = logging.getLogger(".cairodoc")

# ------------------------------------------------------------------------
#
# Constants
#
# ------------------------------------------------------------------------

# resolution
DPI = 72.0


# ------------------------------------------------------------------------
#
# CairoDocgen class
#
# ------------------------------------------------------------------------
class CairoDocgen(libcairodoc.CairoDoc):
    """Render the document into a file using a Cairo surface."""

    def create_cairo_surface(self, fobj, width_in_points, height_in_points):
        # See
        # http://cairographics.org/documentation/pycairo/3/reference/surfaces.html#class-pssurface-surface
        # for the arg semantics.
        raise "Missing surface factory override!!!"

    def run(self):
        """Create the output file.
        The derived class overrides EXT and create_cairo_surface
        """
        # get paper dimensions
        paper_width = self.paper.get_size().get_width() * DPI / 2.54
        paper_height = self.paper.get_size().get_height() * DPI / 2.54
        page_width = round(self.paper.get_usable_width() * DPI / 2.54)
        page_height = round(self.paper.get_usable_height() * DPI / 2.54)
        left_margin = self.paper.get_left_margin() * DPI / 2.54
        top_margin = self.paper.get_top_margin() * DPI / 2.54

        # create cairo context and pango layout
        filename = self._backend.filename
        # Cairo can't reliably handle unicode filenames on Linux or
        # Windows, so open the file for it.
        with open(filename, "wb") as fd:
            try:
                surface = self.create_cairo_surface(fd, paper_width, paper_height)
                surface.set_fallback_resolution(300, 300)
                cr = cairo.Context(surface)
                fontmap = PangoCairo.font_map_new()
                fontmap.set_resolution(DPI)
                pango_context = fontmap.create_context()
                pango_context.set_round_glyph_positions(False)
                options = cairo.FontOptions()
                options.set_hint_metrics(cairo.HINT_METRICS_OFF)
                if is_quartz():
                    PangoCairo.context_set_resolution(pango_context, 72)
                PangoCairo.context_set_font_options(pango_context, options)
                layout = Pango.Layout(pango_context)
                PangoCairo.update_context(cr, pango_context)
                # paginate the document
                self.paginate_document(layout, page_width, page_height, DPI, DPI)
                body_pages = self._pages

                # build the table of contents and alphabetical index
                toc_page = None
                index_page = None
                toc = []
                index = {}
                for page_nr, page in enumerate(body_pages):
                    if page.has_toc():
                        toc_page = page_nr
                    if page.has_index():
                        index_page = page_nr
                    for mark in page.get_marks():
                        if mark.type == INDEX_TYPE_ALP:
                            if mark.key in index:
                                if page_nr + 1 not in index[mark.key]:
                                    index[mark.key].append(page_nr + 1)
                            else:
                                index[mark.key] = [page_nr + 1]
                        elif mark.type == INDEX_TYPE_TOC:
                            toc.append([mark, page_nr + 1])

                # paginate the table of contents
                rebuild_required = False
                if toc_page is not None:
                    toc_pages = self.__generate_toc(
                        layout, page_width, page_height, toc
                    )
                    offset = len(toc_pages) - 1
                    if offset > 0:
                        self.__increment_pages(toc, index, toc_page, offset)
                        rebuild_required = True
                    if index_page and toc_page < index_page:
                        index_page += offset
                else:
                    toc_pages = []

                # paginate the index
                if index_page is not None:
                    index_pages = self.__generate_index(
                        layout, page_width, page_height, index
                    )
                    offset = len(index_pages) - 1
                    if offset > 0:
                        self.__increment_pages(toc, index, index_page, offset)
                        rebuild_required = True
                    if toc_page and toc_page > index_page:
                        toc_page += offset
                else:
                    index_pages = []

                # rebuild the table of contents and index if required
                if rebuild_required:
                    if toc_page is not None:
                        toc_pages = self.__generate_toc(
                            layout, page_width, page_height, toc
                        )
                    if index_page is not None:
                        index_pages = self.__generate_index(
                            layout, page_width, page_height, index
                        )

                # render the pages
                if toc_page is not None:
                    body_pages = (
                        body_pages[:toc_page] + toc_pages + body_pages[toc_page + 1 :]
                    )
                if index_page is not None:
                    body_pages = (
                        body_pages[:index_page]
                        + index_pages
                        + body_pages[index_page + 1 :]
                    )
                self._pages = body_pages
                for page_nr, page in enumerate(self._pages):
                    cr.save()
                    cr.translate(left_margin, top_margin)
                    self.draw_page(
                        page_nr, cr, layout, page_width, page_height, DPI, DPI
                    )
                    cr.show_page()
                    cr.restore()

                # close the surface (file)
                surface.finish()

            except IOError as msg:
                errmsg = "%s\n%s" % (_("Could not create %s") % filename, msg)
                raise ReportError(errmsg)
            except Exception as err:
                errmsg = "%s\n%s" % (_("Could not create %s") % filename, err)
                raise ReportError(errmsg)

    def __increment_pages(self, toc, index, start_page, offset):
        """
        Increment the page numbers in the table of contents and index.
        """
        start_page += 1  # toc/index numbers start at 1
        for n, value in enumerate(toc):
            page_nr = toc[n][1]
            toc[n][1] = page_nr + (offset if page_nr > start_page else 0)
        for key, value in index.items():
            index[key] = [
                page_nr + (offset if page_nr > start_page else 0) for page_nr in value
            ]

    def __generate_toc(self, layout, page_width, page_height, toc):
        """
        Generate the table of contents.
        """
        self._doc = libcairodoc.GtkDocDocument()
        self._active_element = self._doc
        self._pages = []
        write_toc(toc, self)
        self.paginate_document(layout, page_width, page_height, DPI, DPI)
        return self._pages

    def __generate_index(self, layout, page_width, page_height, index):
        """
        Generate the index.
        """
        self._doc = libcairodoc.GtkDocDocument()
        self._active_element = self._doc
        self._pages = []
        write_index(index, self)
        self.paginate_document(layout, page_width, page_height, DPI, DPI)
        return self._pages


def write_toc(toc, doc):
    """
    Write the table of contents.
    """
    if not toc:
        return

    doc.start_paragraph("TOC-Title")
    doc.write_text(doc.toc_title)
    doc.end_paragraph()

    doc.start_table("toc", "TOC-Table")
    for mark, page_nr in toc:
        doc.start_row()
        doc.start_cell("TOC-Cell")
        if mark.level == 1:
            style_name = "TOC-Heading1"
        elif mark.level == 2:
            style_name = "TOC-Heading2"
        else:
            style_name = "TOC-Heading3"
        doc.start_paragraph(style_name)
        doc.write_text(mark.key)
        doc.end_paragraph()
        doc.end_cell()
        doc.start_cell("TOC-Cell")
        doc.start_paragraph(style_name)
        doc.write_text(str(page_nr))
        doc.end_paragraph()
        doc.end_cell()
        doc.end_row()
    doc.end_table()


def write_index(index, doc):
    """
    Write the alphabetical index.
    """
    if not index:
        return

    doc.start_paragraph("IDX-Title")
    doc.write_text(doc.index_title)
    doc.end_paragraph()

    doc.start_table("index", "IDX-Table")
    for key in sorted(index.keys()):
        doc.start_row()
        doc.start_cell("IDX-Cell")
        doc.start_paragraph("IDX-Entry")
        doc.write_text(key)
        doc.end_paragraph()
        doc.end_cell()
        doc.start_cell("IDX-Cell")
        doc.start_paragraph("IDX-Entry")
        pages = [str(page_nr) for page_nr in index[key]]
        # TODO for Arabic, should the next line's comma be translated?
        doc.write_text(", ".join(pages))
        doc.end_paragraph()
        doc.end_cell()
        doc.end_row()
    doc.end_table()


# ------------------------------------------------------------------------
#
# PdfDoc class
#
# ------------------------------------------------------------------------
class PdfDoc(CairoDocgen):
    """Render the document into PDF file using Cairo."""

    def create_cairo_surface(self, fobj, width_in_points, height_in_points):
        return cairo.PDFSurface(fobj, width_in_points, height_in_points)


# ------------------------------------------------------------------------
#
# PsDoc class
#
# ------------------------------------------------------------------------
class PsDoc(CairoDocgen):
    """Render the document into PS file using Cairo."""

    EXT = "ps"

    def create_cairo_surface(self, fobj, width_in_points, height_in_points):
        return cairo.PSSurface(fobj, width_in_points, height_in_points)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Note types.
"""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from .grampstype import GrampsType
from ..const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext


class NoteType(GrampsType):
    UNKNOWN = -1
    CUSTOM = 0
    GENERAL = 1
    RESEARCH = 2
    TRANSCRIPT = 3
    # per object with notes a Type to distinguish the notes
    PERSON = 4
    ATTRIBUTE = 5
    ADDRESS = 6
    ASSOCIATION = 7
    LDS = 8
    FAMILY = 9
    EVENT = 10
    EVENTREF = 11
    SOURCE = 12
    SOURCEREF = 13
    PLACE = 14
    REPO = 15
    REPOREF = 16
    MEDIA = 17
    MEDIAREF = 18
    CHILDREF = 19
    PERSONNAME = 20
    # other common types
    SOURCE_TEXT = 21  # this is used for verbatim source text in SourceRef
    CITATION = 22
    REPORT_TEXT = 23  # this is used for notes used for reports
    # indicate a note is html code
    HTML_CODE = 24
    TODO = 25
    # indicate a note used as link in another note
    LINK = 26
    ANALYSIS = 27

    _CUSTOM = CUSTOM
    _DEFAULT = GENERAL

    _DATAMAPREAL = [
        (UNKNOWN, _("Unknown"), "Unknown"),
        (CUSTOM, _("Custom"), "Custom"),
        (GENERAL, _("General"), "General"),
        (RESEARCH, _("Research"), "Research"),
        (ANALYSIS, _("Analysis"), "Analysis"),
        (TRANSCRIPT, _("Transcript"), "Transcript"),
        (SOURCE_TEXT, _("Source text"), "Source text"),
        (CITATION, _("Citation"), "Citation"),
        (REPORT_TEXT, _("Report"), "Report"),
        (HTML_CODE, _("Html code"), "Html code"),
        (TODO, _("To Do", "notetype"), "To Do"),
        (LINK, _("Link", "notetype"), "Link"),
    ]

    _DATAMAPIGNORE = [
        (PERSON, _("Person Note"), "Person Note"),
        (PERSONNAME, _("Name Note"), "Name Note"),
        (ATTRIBUTE, _("Attribute Note"), "Attribute Note"),
        (ADDRESS, _("Address Note"), "Address Note"),
        (ASSOCIATION, _("Association Note"), "Association Note"),
        (LDS, _("LDS Note"), "LDS Note"),
        (FAMILY, _("Family Note"), "Family Note"),
        (EVENT, _("Event Note"), "Event Note"),
        (EVENTREF, _("Event Reference Note"), "Event Reference Note"),
        (SOURCE, _("Source Note"), "Source Note"),
        (SOURCEREF, _("Source Reference Note"), "Source Reference Note"),
        (PLACE, _("Place Note"), "Place Note"),
        (REPO, _("Repository Note"), "Repository Note"),
        (REPOREF, _("Repository Reference Note"), "Repository Reference Note"),
        (MEDIA, _("Media Note"), "Media Note"),
        (MEDIAREF, _("Media Reference Note"), "Media Reference Note"),
        (CHILDREF, _("Child Reference Note"), "Child Reference Note"),
    ]

    _DATAMAP = _DATAMAPREAL + _DATAMAPIGNORE

    def __init__(self, value=None):
        GrampsType.__init__(self, value)

    def get_ignore_list(self, exception):
        """
        Return a list of the types to ignore and not include in default lists.

        Exception is a sublist of types that may not be ignored

        :param exception: list of integer values corresponding with types that
                          have to be excluded from the ignore list
        :type exception: list
        :returns: list of integers corresponding with the types to ignore when
                  showing a list of different types
        :rtype: list

        """
        ignlist = [x[0] for x in self._DATAMAPIGNORE]
        if exception:
            for type_ in exception:
                try:
                    del ignlist[ignlist.index(type_)]
                except ValueError:
                    pass
        return ignlist



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Package providing filtering framework for Gramps.
"""

CustomFilters = None

from ..const import CUSTOM_FILTERS
from ._filterlist import FilterList
from ._genericfilter import (
    GenericFilter,
    GenericFilterFactory,
    DeferredFilter,
    DeferredFamilyFilter,
)
from ._paramfilter import ParamFilter
from ._searchfilter import SearchFilter, ExactSearchFilter


def reload_custom_filters():
    global CustomFilters
    CustomFilters = FilterList(CUSTOM_FILTERS)
    CustomFilters.load()


# if not CustomFilters:  # moved to viewmanager
# reload_custom_filters()



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2013       Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""Descendant Gramplet"""

# -------------------------------------------------------------------------
#
# GTK modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk

# ------------------------------------------------------------------------
#
# Gramps modules
#
# ------------------------------------------------------------------------
from gramps.gen.plug import Gramplet
from gramps.gui.listmodel import ListModel, NOSORT
from gramps.gui.editors import EditPerson
from gramps.gen.errors import WindowActiveError
from gramps.gen.display.name import displayer as name_displayer
from gramps.gen.datehandler import get_date
from gramps.gen.utils.db import get_birth_or_fallback, get_death_or_fallback
from gramps.gui.utils import model_to_text, text_to_clipboard
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext


class Descendant(Gramplet):
    """
    Gramplet to display descendants of the active person.
    """

    def init(self):
        self.gui.WIDGET = self.build_gui()
        self.gui.get_container_widget().remove(self.gui.textview)
        self.gui.get_container_widget().add(self.gui.WIDGET)
        self.gui.WIDGET.show()

    def build_gui(self):
        """
        Build the GUI interface.
        """
        self.view = Gtk.TreeView()
        self.view.set_tooltip_column(3)
        titles = [
            (_("Name"), 0, 230),
            (_("Birth"), 2, 100),
            ("", NOSORT, 1),
            ("", NOSORT, 1),  # tooltip
            ("", NOSORT, 100),
        ]  # handle
        self.model = ListModel(
            self.view,
            titles,
            list_mode="tree",
            event_func=self.cb_double_click,
            right_click=self.cb_right_click,
        )
        return self.view

    def get_has_data(self, active_handle):
        """
        Return True if the gramplet has data, else return False.
        """
        if active_handle:
            person = self.dbstate.db.get_person_from_handle(active_handle)
            if person:
                for family_handle in person.get_family_handle_list():
                    family = self.dbstate.db.get_family_from_handle(family_handle)
                    if family:
                        for child_ref in family.get_child_ref_list():
                            return True
        return False

    def cb_double_click(self, treeview):
        """
        Handle double click on treeview.
        """
        (model, iter_) = treeview.get_selection().get_selected()
        if not iter_:
            return

        try:
            handle = model.get_value(iter_, 4)
            person = self.dbstate.db.get_person_from_handle(handle)
            EditPerson(self.dbstate, self.uistate, [], person)
        except WindowActiveError:
            pass

    def cb_right_click(self, treeview, event):
        """
        Handle right click on treeview.
        """
        (model, iter_) = treeview.get_selection().get_selected()
        sensitivity = 1 if iter_ else 0
        menu = Gtk.Menu()
        menu.set_reserve_toggle_size(False)
        entries = [
            (_("Edit"), lambda obj: self.cb_double_click(treeview), sensitivity),
            (None, None, 0),
            (_("Copy all"), lambda obj: self.on_copy_all(treeview), 1),
        ]
        for title, callback, sensitivity in entries:
            item = Gtk.MenuItem(label=title)
            if callback:
                item.connect("activate", callback)
            else:
                item = Gtk.SeparatorMenuItem()
            item.set_sensitive(sensitivity)
            item.show()
            menu.append(item)
        self.menu = menu
        self.menu.popup_at_pointer(event)

    def on_copy_all(self, treeview):
        """
        Copy tree to clipboard.
        """
        model = treeview.get_model()
        text = model_to_text(model, [0, 1], level=1)
        text_to_clipboard(text)

    def db_changed(self):
        self.update()

    def active_changed(self, handle):
        self.update()

    def update_has_data(self):
        active_handle = self.get_active("Person")
        if active_handle:
            self.set_has_data(self.get_has_data(active_handle))
        else:
            self.set_has_data(False)

    def main(self):
        active_handle = self.get_active("Person")
        self.model.clear()
        if active_handle:
            self.add_to_tree(None, active_handle)
            self.view.expand_all()
            self.set_has_data(self.get_has_data(active_handle))
        else:
            self.set_has_data(False)

    def add_to_tree(self, parent_id, person_handle):
        """
        Add a person to the tree.
        """
        person = self.dbstate.db.get_person_from_handle(person_handle)
        name = name_displayer.display(person)

        birth = get_birth_or_fallback(self.dbstate.db, person)
        death = get_death_or_fallback(self.dbstate.db, person)

        birth_text = birth_date = birth_sort = ""
        if birth:
            birth_date = get_date(birth)
            birth_sort = "%012d" % birth.get_date_object().get_sort_value()
            birth_text = _("%(abbr)s %(date)s") % {
                "abbr": birth.type.get_abbreviation(),
                "date": birth_date,
            }

        death_date = death_sort = death_text = ""
        if death:
            death_date = get_date(death)
            death_sort = "%012d" % death.get_date_object().get_sort_value()
            death_text = _("%(abbr)s %(date)s") % {
                "abbr": death.type.get_abbreviation(),
                "date": death_date,
            }

        tooltip = name + "\n" + birth_text + "\n" + death_text

        item_id = self.model.add(
            [name, birth_date, birth_sort, tooltip, person_handle], node=parent_id
        )

        for family_handle in person.get_family_handle_list():
            family = self.dbstate.db.get_family_from_handle(family_handle)
            for child_ref in family.get_child_ref_list():
                self.add_to_tree(item_id, child_ref.ref)
        return item_id



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2008  Gary Burton
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Standard Python modules
#
# -------------------------------------------------------------------------
from ....const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ....datehandler import parser
from ....display.place import displayer as place_displayer
from ....lib.eventtype import EventType
from .. import Rule


# -------------------------------------------------------------------------
#
# HasBirth
#
# -------------------------------------------------------------------------
class HasData(Rule):
    """Rule that checks for an event containing particular values"""

    labels = [_("Event type:"), _("Date:"), _("Place:"), _("Description:")]
    name = _("Events with <data>")
    description = _("Matches events with data of a particular value")
    category = _("General filters")
    allow_regex = True

    def prepare(self, db, user):
        self.event_type = self.list[0]
        self.date = self.list[1]

        if self.event_type:
            self.event_type = EventType()
            self.event_type.set_from_xml_str(self.list[0])

        if self.date:
            self.date = parser.parse(self.date)

    def apply(self, db, event):
        if self.event_type and event.get_type() != self.event_type:
            # No match
            return False

        if self.date and not event.get_date_object().match(self.date):
            # No match
            return False

        if self.list[2]:
            place_id = event.get_place_handle()
            if place_id:
                place = db.get_place_from_handle(place_id)
                place_title = place_displayer.display(db, place)
                if not self.match_substring(2, place_title):
                    # No match
                    return False
            else:
                # No place attached to event
                return False

        if not self.match_substring(3, event.get_description()):
            # No match
            return False

        # All conditions matched
        return True



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2010       Michiel D. Nauta
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
NoteBase class for Gramps.
"""
# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
import logging

LOG = logging.getLogger(".note")


# -------------------------------------------------------------------------
#
# NoteBase class
#
# -------------------------------------------------------------------------
class NoteBase:
    """
    Base class for storing notes.

    Starting in 3.0 branch, the objects may have multiple notes.
    Internally, this class maintains a list of Note handles,
    as a note_list attribute of the NoteBase object.
    """

    def __init__(self, source=None):
        """
        Create a new NoteBase, copying from source if not None.

        :param source: Object used to initialize the new object
        :type source: NoteBase
        """
        self.note_list = list(source.note_list) if source else []

    def serialize(self):
        """
        Convert the object to a serialized tuple of data.
        """
        return self.note_list

    def unserialize(self, data):
        """
        Convert a serialized tuple of data to an object.
        """
        self.note_list = list(data)

    def add_note(self, handle):
        """
        Add the :class:`~.note.Note` handle to the list of note handles.

        :param handle: :class:`~.note.Note` handle to add the list of notes
        :type handle: str

        :returns: True if handle was added, False if it already was in the list
        :rtype: bool
        """
        if handle in self.note_list:
            return False
        else:
            self.note_list.append(handle)
            return True

    def remove_note(self, handle):
        """
        Remove the specified handle from the list of note handles, and all
        secondary child objects.

        :param handle: :class:`~.note.Note` handle to remove from the list of
                       notes
        :type handle: str
        """
        if handle in self.note_list:
            self.note_list.remove(handle)
        for item in self.get_note_child_list():
            item.remove_note(handle)

    def get_note_child_list(self):
        """
        Return the list of child secondary objects that may refer notes.

        All methods which inherit from NoteBase and have other child objects
        with notes, should return here a list of child objects which are
        NoteBase

        :returns: Returns the list of child secondary child objects that may
                refer notes.
        :rtype: list
        """
        return []

    def get_note_list(self):
        """
        Return the list of :class:`~.note.Note` handles associated with the
        object.

        :returns: The list of :class:`~.note.Note` handles
        :rtype: list
        """
        return self.note_list

    def has_note_reference(self, note_handle):
        """
        Return True if the object or any of its child objects has reference
        to this note handle.

        :param note_handle: The note handle to be checked.
        :type note_handle: str
        :returns: Returns whether the object or any of its child objects has
                  reference to this note handle.
        :rtype: bool
        """
        for note_ref in self.note_list:
            if note_ref == note_handle:
                return True

        for item in self.get_note_child_list():
            if item.has_note_reference(note_handle):
                return True

        return False

    def remove_note_references(self, handle_list):
        """
        Remove the specified handles from the list of note handles, and all
        secondary child objects.

        :param citation_handle_list: The list of note handles to be removed
        :type handle: list
        """
        LOG.debug(
            "enter remove_note handle: %s self: %s note_list: %s",
            handle_list,
            self,
            self.note_list,
        )
        for handle in handle_list:
            if handle in self.note_list:
                LOG.debug("remove handle %s from note_list %s", handle, self.note_list)
                self.note_list.remove(handle)
        LOG.debug("get_note_child_list %s", self.get_note_child_list())
        for item in self.get_note_child_list():
            item.remove_note_references(handle_list)

    def set_note_list(self, note_list):
        """
        Assign the passed list to be object's list of :class:`~.note.Note`
        handles.

        :param note_list: List of :class:`~.note.Note` handles to be set on the
                          object
        :type note_list: list
        """
        self.note_list = note_list

    def _merge_note_list(self, acquisition):
        """
        Merge the list of notes from acquisition with our own.

        :param acquisition: The note list of this object will be merged with
                            the current note list.
        :rtype acquisition: NoteBase
        """
        for addendum in acquisition.note_list:
            self.add_note(addendum)

    def get_referenced_note_handles(self):
        """
        Return the list of (classname, handle) tuples for all referenced notes.

        This method should be used to get the :class:`~.note.Note` portion of
        the list by objects that store note lists.

        :returns: List of (classname, handle) tuples for referenced objects.
        :rtype: list
        """
        return [("Note", handle) for handle in self.note_list]

    def replace_note_references(self, old_handle, new_handle):
        """
        Replace references to note handles in the list of this object and
        all child objects and merge equivalent entries.

        :param old_handle: The note handle to be replaced.
        :type old_handle: str
        :param new_handle: The note handle to replace the old one with.
        :type new_handle: str
        """
        refs_list = self.note_list[:]
        new_ref = None
        if new_handle in self.note_list:
            new_ref = new_handle
        n_replace = refs_list.count(old_handle)
        for ix_replace in range(n_replace):
            idx = refs_list.index(old_handle)
            if new_ref:
                self.note_list.pop(idx)
                refs_list.pop(idx)
            else:
                self.note_list[idx] = new_handle

        for item in self.get_note_child_list():
            item.replace_note_references(old_handle, new_handle)



# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2009  Benny Malengier
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Package init for the treemodels package.
"""

from .peoplemodel import PeopleBaseModel, PersonListModel, PersonTreeModel
from .familymodel import FamilyModel
from .eventmodel import EventModel
from .sourcemodel import SourceModel
from .placemodel import PlaceBaseModel, PlaceListModel, PlaceTreeModel
from .mediamodel import MediaModel
from .repomodel import RepositoryModel
from .notemodel import NoteModel
from .citationbasemodel import CitationBaseModel
from .citationlistmodel import CitationListModel
from .citationtreemodel import CitationTreeModel
